#!/usr/bin/env python
# coding: utf-8

import time
import pybullet
import threading
import pybullet_data
from qibullet.laser import Laser
from qibullet.camera import Camera
from qibullet.nao_virtual import NaoVirtual
from qibullet.romeo_virtual import RomeoVirtual
from qibullet.pepper_virtual import PepperVirtual
from qibullet.robot_module import RobotModule


class SimulationManager:
    """
    Class allowing to handle the different parameters of a pybullet simulation
    """

    def __init__(self):
        """
        Constructor
        """
        pass

    def launchSimulation(self, gui=True, use_shared_memory=False):
        """
        Launches a simulation instance

        Parameters:
            gui - Boolean, if True the simulation is launched with a GUI, and
            with no GUI otherwise
            use_shared_memory - Experimental feature, only taken into account
            if gui=False, False by default. If True, the simulation will use
            the pybullet SHARED_MEMORY_SERVER mode to create an instance. If
            multiple simulation instances are created, this solution allows a
            multicore parallelisation of the bullet motion servers (one for
            each instance). In DIRECT mode, such a parallelisation is not
            possible and the motion servers are manually stepped using the
            _stepSimulation method. (More information in the setup section of
            the qiBullet wiki, and in the pybullet documentation)

        Returns:
            physics_client - The id of the simulation client created
        """
        if gui:  # pragma: no cover
            physics_client = pybullet.connect(pybullet.GUI)
            pybullet.setRealTimeSimulation(1, physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_RGB_BUFFER_PREVIEW,
                0,
                physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_DEPTH_BUFFER_PREVIEW,
                0,
                physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_SEGMENTATION_MARK_PREVIEW,
                0,
                physicsClientId=physics_client)
        else:
            if use_shared_memory:
                physics_client = pybullet.connect(
                    pybullet.SHARED_MEMORY_SERVER)

                pybullet.setRealTimeSimulation(
                    enableRealTimeSimulation=1,
                    physicsClientId=physics_client)
            else:
                physics_client = pybullet.connect(pybullet.DIRECT)
                threading.Thread(
                    target=self._stepSimulation,
                    args=[physics_client]).start()

        pybullet.setGravity(0, 0, -9.81, physicsClientId=physics_client)
        return physics_client

    def resetSimulation(self, physics_client):
        """
        Resets the simulated instance corresponding to the physics client id.
        All of the objects loaded in the simulation will be destroyed, but the
        instance will still be running
        """
        self._clearInstance(physics_client)
        pybullet.resetSimulation(physicsClientId=physics_client)

    def stopSimulation(self, physics_client):
        """
        Stops the simulated instance corresponding to the physics_client id

        Parameters:
            physics_client - The id of the simulated instance to be stopped
        """
        self._clearInstance(physics_client)

        try:
            pybullet.disconnect(physicsClientId=physics_client)

        except pybullet.error:
            print("Instance " + str(physics_client) + " already stopped")

    def setLightPosition(self, physics_client, light_position):
        """
        Set the position of the GUI's light (does not work in DIRECT mode)

        Parameters:
            light_position - List containing the 3D positions [x, y, z] along
            the X, Y, and Z axis in the world frame, in meters
        """
        try:
            assert isinstance(light_position, list)
            assert len(light_position) == 3

            pybullet.configureDebugVisualizer(
                lightPosition=light_position,
                physicsClientId=physics_client)

        except AssertionError:
            raise pybullet.error("Incorrect light position format")

    def spawnPepper(
            self,
            physics_client,
            translation=[0, 0, 0],
            quaternion=[0, 0, 0, 1],
            spawn_ground_plane=False):
        """
        Loads a Pepper model in the simulation

        Parameters:
            physics_client - The id of the simulated instance in which the
            robot is supposed to be spawned
            translation - List containing 3 elements, the spawning translation
            [x, y, z] in the WORLD frame
            quaternions - List containing 4 elements, the spawning rotation as
            a quaternion [x, y, z, w] in the WORLD frame
            spawn_ground_plane - If True, the pybullet_data ground plane will
            be spawned

        Returns:
            pepper_virtual - A PepperVirtual object, the Pepper simulated
            instance
        """
        pepper_virtual = PepperVirtual()

        if spawn_ground_plane:
            self._spawnGroundPlane(physics_client)

        pepper_virtual.loadRobot(
            translation,
            quaternion,
            physicsClientId=physics_client)

        return pepper_virtual

    def spawnNao(
            self,
            physics_client,
            translation=[0, 0, 0],
            quaternion=[0, 0, 0, 1],
            spawn_ground_plane=False):
        """
        Loads a NAO model in the simulation

        Parameters:
            physics_client - The id of the simulated instance in which the
            robot is supposed to be spawned
            translation - List containing 3 elements, the spawning translation
            [x, y, z] in the WORLD frame
            quaternions - List containing 4 elements, the spawning rotation as
            a quaternion [x, y, z, w] in the WORLD frame
            spawn_ground_plane - If True, the pybullet_data ground plane will
            be spawned

        Returns:
            nao_virtual - A NaoVirtual object, the NAO simulated instance
        """
        nao_virtual = NaoVirtual()

        if spawn_ground_plane:
            self._spawnGroundPlane(physics_client)

        nao_virtual.loadRobot(
            translation,
            quaternion,
            physicsClientId=physics_client)

        return nao_virtual

    def spawnRomeo(
            self,
            physics_client,
            translation=[0, 0, 0],
            quaternion=[0, 0, 0, 1],
            spawn_ground_plane=False):
        """
        Loads a Romeo model in the simulation

        Parameters:
            physics_client - The id of the simulated instance in which the
            robot is supposed to be spawned
            translation - List containing 3 elements, the spawning translation
            [x, y, z] in the WORLD frame
            quaternions - List containing 4 elements, the spawning rotation as
            a quaternion [x, y, z, w] in the WORLD frame
            spawn_ground_plane - If True, the pybullet_data ground plane will
            be spawned

        Returns:
            romeo_virtual - A RomeoVirtual object, the Romeo simulated instance
        """
        romeo_virtual = RomeoVirtual()

        if spawn_ground_plane:
            self._spawnGroundPlane(physics_client)

        romeo_virtual.loadRobot(
            translation,
            quaternion,
            physicsClientId=physics_client)

        return romeo_virtual

    def removePepper(self, pepper_virtual):
        """
        Removes a Pepper from a simulated instance

        Parameters:
            pepper_virtual - The virtual Pepper robot to be removed
        """
        pepper_virtual.laser_manager._terminateModule()
        pepper_virtual.base_controller._terminateModule()
        self._removeRobot(pepper_virtual)

    def removeNao(self, nao_virtual):
        """
        Removes a NAO from a simulated instance

        Parameters:
            nao_virtual - The virtual NAO robot to be removed
        """
        self._removeRobot(nao_virtual)

    def removeRomeo(self, romeo_virtual):
        """
        Removes a Romeo from a simulated instance

        Parameters:
            romeo_virtual - The virtual Romeo to be removed
        """
        self._removeRobot(romeo_virtual)

    def _removeRobot(self, robot_virtual):
        """
        Removes a Virtual robot (Robot inheriting from RobotVirtual) from a
        simulated instance

        Parameters:
            robot_virtual - The virtual robot to be removed
        """
        for camera in robot_virtual.camera_dict.values():
            if id(camera) in Camera._getCameraHandlesDict():
                camera.unsubscribe()

        pybullet.removeBody(robot_virtual.getRobotModel())

    def _clearInstance(self, physics_client):
        """
        INTERNAL METHOD, Called to kill the processes of modules running in a
        simulated instance, before resetting or stopping it.

        Parameters:
            physics_client - The client id of the simulated instance that will
            be cleared
        """
        for module in RobotModule._getInstances():
            if module.getPhysicsClientId() == physics_client:
                module._terminateModule()

        for camera in Camera._getCameraHandlesDict().values():
            if physics_client == camera.getPhysicsClientId():
                camera.unsubscribe()

    def _stepSimulation(self, physics_client):
        """
        INTERNAL METHOD: This method is only used for a simulation in DIRECT
        mode (without the gui).

        Parameters:
            physics_client - The id of the simulated instance to be stepped
        """
        try:
            initial_time = time.time()
            while True:
                pybullet.stepSimulation(physicsClientId=physics_client)
                time.sleep(1./10000.)
        except Exception:
            pass

    def _spawnGroundPlane(self, physics_client):
        """
        INTERNAL METHOD, Loads a ground plane

        Parameters:
            physics_client - The id of the simulated instance in which the
            ground plane is supposed to be spawned
        """
        pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
        pybullet.loadMJCF(
            "mjcf/ground_plane.xml",
            physicsClientId=physics_client)



#!/usr/bin/env python
# coding: utf-8

import atexit
import weakref
import threading


class RobotModule:
    """
    Abstract class describing a robot module. The Sensor and Controller classes
    inherit that class
    """
    _instances = set()

    def __init__(self, robot_model, physics_client):
        """
        Constructor

        Parameters:
            robot_model - The pybullet model of the robot
            physics_client - The if of the simulated instance in which the
            robot is spawned
        """
        self.robot_model = robot_model
        self.physics_client = physics_client
        self.module_process = threading.Thread(target=None)
        self._instances.add(weakref.ref(self))
        self._module_termination = False
        atexit.register(self._terminateModule)

    def getRobotModel(self):
        """
        Returns the pybullet model to which the module is associated

        Returns:
            robot_model - The pybullet model of the robot
        """
        return self.robot_model

    def getPhysicsClientId(self):
        """
        Returns the id of the simulated instance in which the module is loaded

        Returns:
            physics_client - The id of the simulation in which the robot
            (possessing the module) is spawned
        """
        return self.physics_client

    @classmethod
    def _getInstances(cls):
        """
        INTERNAL CLASSMETHOD, get all of the RobotModule (and daughters)
        instances
        """
        dead = set()

        for ref in cls._instances:
            obj = ref()

            if obj is not None:
                yield obj
            else:
                dead.add(ref)

        cls._instances -= dead

    def _terminateModule(self):
        """
        INTERNAL METHOD, can be called to terminate an asynchronous module
        process. Should only be used when killing the simulation
        """
        self._module_termination = True

        if self.module_process.isAlive():
            self.module_process.join()



#!/usr/bin/env python
# coding: utf-8

import time
import math
import pybullet
import threading
from qibullet.sensor import Sensor

RAY_MISS_COLOR = [0, 1, 0]
RAY_HIT_COLOR = [1, 0, 0]
NUM_RAY = 15
RAY_LENGTH = 3.0  # The theoretical length is 5.6, closer to 3.0 in reality
LASER_ANGLE = 60
LASER_POSITION = [
    [0.0562, 0, -0.334],  # Front laser
    [-0.018, -0.0899, -0.334],  # Right laser
    [-0.018, 0.0899, -0.334]  # Left laser
]
ANGLE_LIST_POSITION = [
    math.radians(LASER_ANGLE/2),  # Front laser
    math.radians(LASER_ANGLE/2) - 1.75728,  # Right laser
    math.radians(LASER_ANGLE/2) + 1.75728  # Left laser
]

NUM_LASER = len(LASER_POSITION)
LASER_FRAMERATE = 6.25


class Laser(Sensor):
    """
    Class representing a virtual laser
    """

    def __init__(
            self,
            robot_model,
            laser_id,
            physicsClientId=0,
            display=False):
        """
        Constructor

        Parameters:
            robot_model - The pybullet model of the robot.
            laser_id - The id of the link (Link type)
            onto which the Lasers' are attached.
            physicsClientId - The id of the simulated instance in which the
            lasers are to be spawned
            display - boolean that allow the display of the laser
        """
        Sensor.__init__(self, robot_model, physicsClientId)
        self.ray_from = []
        self.ray_to = []
        self.ray_ids = []
        self.laser_value = [0] * NUM_RAY * NUM_LASER
        self.laser_id = laser_id
        self.display = display

    def isActive(self):
        """
        Check if the lasers are subscribed or not (if the laser thread process
        is active or not)

        Returns:
            boolean - True if the lasers are subscribed, false otherwise
        """
        return self.module_process.isAlive()

    def subscribe(self):
        """
        Subscribe to the laser scan (this will activate the laser scan
        process).
        """
        # No need to subscribe to the laser scan if the lasers are activated
        if self.isActive():
            return

        self._module_termination = False
        self._initializeRays()
        self.module_process = threading.Thread(target=self._laserScan)
        self.module_process.start()

    def unsubscribe(self):
        """
        Unsubscribe from the laser scan (this will deactivate the laser scan
        process)
        """
        if self.isActive():
            self._terminateModule()

    def showLaser(self, display):
        """
        Display debug lines that simulate the laser
        """
        self.display = display

    def getFrontLaserValue(self):
        """
        Return a list of the front laser value (clockwise)
        """
        return self.laser_value[:NUM_RAY]

    def getRightLaserValue(self):
        """
        Return a list of the right laser value (clockwise)
        """
        return self.laser_value[NUM_RAY:2*NUM_RAY]

    def getLeftLaserValue(self):
        """
        Return a list of the left laser value (clockwise)
        """
        return self.laser_value[2*NUM_RAY:]

    def _initializeRays(self):
        """
        INTERNAL METHOD, initialize the laser and all variables needed
        """
        for index in range(NUM_LASER):
            angle = ANGLE_LIST_POSITION[index]
            for i in range(NUM_RAY):
                self.ray_from.append([
                    LASER_POSITION[index][0],
                    LASER_POSITION[index][1],
                    LASER_POSITION[index][2]])
                self.ray_to.append(
                    [LASER_POSITION[index][0] + (RAY_LENGTH) *
                     math.cos(float(i) *
                     math.radians(-LASER_ANGLE)/NUM_RAY + angle),
                     LASER_POSITION[index][1] + (RAY_LENGTH) *
                     math.sin(float(i) *
                     math.radians(-LASER_ANGLE)/NUM_RAY + angle),
                     LASER_POSITION[index][2]])

    def _laserScan(self):
        """
        INTERNAL METHOD, a loop that simulate the laser and update the distance
        value of each laser
        """
        lastLidarTime = time.time()

        while not self._module_termination:
            nowLidarTime = time.time()

            if (nowLidarTime-lastLidarTime > 1/LASER_FRAMERATE):
                results = pybullet.rayTestBatch(
                    self.ray_from,
                    self.ray_to,
                    parentObjectUniqueId=self.robot_model,
                    parentLinkIndex=self.laser_id,
                    physicsClientId=self.physics_client)

                for i in range(NUM_RAY*len(ANGLE_LIST_POSITION)):
                    hitObjectUid = results[i][0]
                    hitFraction = results[i][2]
                    hitPosition = results[i][3]
                    self.laser_value[i] = hitFraction * RAY_LENGTH

                    if self.display:
                        if not self.ray_ids:
                            self._createDebugLine()

                        if (hitFraction == 1.):
                            pybullet.addUserDebugLine(
                                self.ray_from[i],
                                self.ray_to[i],
                                RAY_MISS_COLOR,
                                replaceItemUniqueId=self.ray_ids[i],
                                parentObjectUniqueId=self.robot_model,
                                parentLinkIndex=self.laser_id,
                                physicsClientId=self.physics_client)
                        else:  # pragma: no cover
                            localHitTo = [self.ray_from[i][0]+hitFraction*(
                                        self.ray_to[i][0]-self.ray_from[i][0]),
                                         self.ray_from[i][1]+hitFraction*(
                                        self.ray_to[i][1]-self.ray_from[i][1]),
                                         self.ray_from[i][2]+hitFraction*(
                                        self.ray_to[i][2]-self.ray_from[i][2])]
                            pybullet.addUserDebugLine(
                                self.ray_from[i],
                                localHitTo,
                                RAY_HIT_COLOR,
                                replaceItemUniqueId=self.ray_ids[i],
                                parentObjectUniqueId=self.robot_model,
                                parentLinkIndex=self.laser_id,
                                physicsClientId=self.physics_client)

                    else:
                        if self.ray_ids:
                            self._resetDebugLine()

                lastLidarTime = nowLidarTime

    def _createDebugLine(self):
        """
        INTERNAL METHOD, create all debug lines needed for simulating the
        lasers
        """
        for i in range(NUM_RAY * NUM_LASER):
            self.ray_ids.append(pybullet.addUserDebugLine(
                self.ray_from[i],
                self.ray_to[i],
                RAY_MISS_COLOR,
                parentObjectUniqueId=self.robot_model,
                parentLinkIndex=self.laser_id,
                physicsClientId=self.physics_client))

    def _resetDebugLine(self):
        """
        INTERNAL METHOD, remove all debug lines
        """
        for i in range(len(self.ray_ids)):
            pybullet.removeUserDebugItem(self.ray_ids[i], self.physics_client)

        self.ray_ids = []



#!/usr/bin/env python
# coding: utf-8

import os
import sys
import atexit
import pybullet
from qibullet.camera import Camera
from qibullet.camera import CameraRgb
from qibullet.camera import CameraDepth
from qibullet.nao_virtual import NaoVirtual
from qibullet.romeo_virtual import RomeoVirtual
from qibullet.pepper_virtual import PepperVirtual
from qibullet.base_controller import PepperBaseController
from threading import Thread

try:
    import rospy
    import roslib
    import roslaunch
    import tf2_ros
    from cv_bridge import CvBridge
    from sensor_msgs.msg import Image
    from sensor_msgs.msg import CameraInfo
    from sensor_msgs.msg import JointState
    from sensor_msgs.msg import LaserScan
    from std_msgs.msg import Header
    from std_msgs.msg import Empty
    from naoqi_bridge_msgs.msg import JointAnglesWithSpeed
    from geometry_msgs.msg import TransformStamped
    from geometry_msgs.msg import Twist
    from nav_msgs.msg import Odometry

    try:
        from naoqi_bridge_msgs.msg import PoseStampedWithSpeed as MovetoPose
        OFFICIAL_DRIVER = False
        print("Using softbankrobotics-research forked version of NAOqi driver")

    except ImportError as e:
        from geometry_msgs.msg import PoseStamped as MovetoPose
        OFFICIAL_DRIVER = True

    MISSING_IMPORT = None

except ImportError as e:
    MISSING_IMPORT = str(e)

TOP_OPTICAL_FRAME = "CameraTop_optical_frame"
BOTTOM_OPTICAL_FRAME = "CameraBottom_optical_frame"
DEPTH_OPTICAL_FRAME = "CameraDepth_optical_frame"


class RosWrapper:
    """
    Virtual class defining the basis of a robot ROS wrapper
    """

    def __init__(self):
        """
        Constructor
        """
        if MISSING_IMPORT is not None:
            raise pybullet.error(MISSING_IMPORT)

        self.spin_thread = None
        self._wrapper_termination = False
        self.image_bridge = CvBridge()
        self.front_info_msg = dict()
        self.bottom_info_msg = dict()
        self.depth_info_msg = dict()
        self.roslauncher = None
        self.transform_broadcaster = tf2_ros.TransformBroadcaster()
        atexit.register(self.stopWrapper)

    def stopWrapper(self):
        """
        Stops the ROS wrapper
        """
        self._wrapper_termination = True

        try:
            assert self.spin_thread is not None
            assert isinstance(self.spin_thread, Thread)
            assert self.spin_thread.isAlive()
            self.spin_thread.join()

        except AssertionError:
            pass

        if self.roslauncher is not None:
            self.roslauncher.stop()
            print("Stopping roslauncher")

    def launchWrapper(self, virtual_robot, ros_namespace, frequency=200):
        """
        Launches the ROS wrapper

        Parameters:
            virtual_robot - The instance of the simulated model
            ros_namespace - The ROS namespace to be added before the ROS topics
            advertized and subscribed
            frequency - The frequency of the ROS rate that will be used to pace
            the wrapper's main loop
        """
        if MISSING_IMPORT is not None:
            raise pybullet.error(MISSING_IMPORT)

        self.robot = virtual_robot
        self.ros_namespace = ros_namespace
        self.frequency = frequency

        rospy.init_node(
            "qibullet_wrapper",
            anonymous=True,
            disable_signals=False)

        # Upload the robot description to the ros parameter server
        try:
            if isinstance(self.robot, PepperVirtual):
                robot_name = "pepper"
            elif isinstance(self.robot, NaoVirtual):
                robot_name = "nao"
            elif isinstance(self.robot, RomeoVirtual):
                robot_name = "romeo"
            else:
                raise pybullet.error(
                    "Unknown robot type, wont set robot description")

            package_path = roslib.packages.get_pkg_dir("naoqi_driver")
            urdf_path = package_path + "/share/urdf/" + robot_name + ".urdf"

            with open(urdf_path, 'r') as file:
                robot_description = file.read()

            rospy.set_param("/robot_description", robot_description)

        except IOError as e:
            raise pybullet.error(
                "Could not retrieve robot descrition: " + str(e))

        # Launch the robot state publisher
        robot_state_publisher = roslaunch.core.Node(
            "robot_state_publisher",
            "robot_state_publisher")

        self.roslauncher = roslaunch.scriptapi.ROSLaunch()
        self.roslauncher.start()
        self.roslauncher.launch(robot_state_publisher)

        # Initialize the ROS publisher and subscribers
        self._initPublishers()
        self._initSubscribers()

        # Launch the wrapper's main loop
        self._wrapper_termination = False
        self.spin_thread = Thread(target=self._spin)
        self.spin_thread.start()

    def _initPublishers(self):
        """
        ABSTRACT INTERNAL METHOD, needs to be implemented in each daughter
        class. Initializes the ROS publishers
        """
        raise NotImplementedError

    def _initSubscribers(self):
        """
        ABSTRACT INTERNAL METHOD, needs to be implemented in each daughter
        class. Initializes the ROS subscribers
        """
        raise NotImplementedError

    def _spin(self):
        """
        ABSTRACT INTERNAL METHOD, needs to be implemented in each daughter
        class. Designed to emulate a ROS spin method
        """
        raise NotImplementedError

    def _broadcastOdometry(self, odometry_publisher):
        """
        INTERNAL METHOD, computes an odometry message based on the robot's
        position, and broadcast it

        Parameters:
            odometry_publisher - The ROS publisher for the odometry message
        """
        # Send Transform odom
        x, y, theta = self.robot.getPosition()
        odom_trans = TransformStamped()
        odom_trans.header.frame_id = "odom"
        odom_trans.child_frame_id = "base_link"
        odom_trans.header.stamp = rospy.get_rostime()
        odom_trans.transform.translation.x = x
        odom_trans.transform.translation.y = y
        odom_trans.transform.translation.z = 0
        quaternion = pybullet.getQuaternionFromEuler([0, 0, theta])
        odom_trans.transform.rotation.x = quaternion[0]
        odom_trans.transform.rotation.y = quaternion[1]
        odom_trans.transform.rotation.z = quaternion[2]
        odom_trans.transform.rotation.w = quaternion[3]
        self.transform_broadcaster.sendTransform(odom_trans)
        # Set up the odometry
        odom = Odometry()
        odom.header.stamp = rospy.get_rostime()
        odom.header.frame_id = "odom"
        odom.pose.pose.position.x = x
        odom.pose.pose.position.y = y
        odom.pose.pose.position.z = 0.0
        odom.pose.pose.orientation = odom_trans.transform.rotation
        odom.child_frame_id = "base_link"
        [vx, vy, vz], [wx, wy, wz] = pybullet.getBaseVelocity(
            self.robot.getRobotModel(),
            self.robot.getPhysicsClientId())
        odom.twist.twist.linear.x = vx
        odom.twist.twist.linear.y = vy
        odom.twist.twist.angular.z = wz
        odometry_publisher.publish(odom)

    def _broadcastCamera(self, camera, image_publisher, info_publisher):
        """
        INTERNAL METHOD, computes the image message and the info message of the
        given camera and publishes them into the ROS framework

        Parameters:
            camera - The camera used for broadcasting
            image_publisher - The ROS publisher for the Image message,
            corresponding to the image delivered by the active camera
            info_publisher - The ROS publisher for the CameraInfo message,
            corresponding to the parameters of the active camera
        """
        try:
            frame = camera.getFrame()
            assert frame is not None

            # Fill the camera info message
            info_msg = CameraInfo()
            info_msg.distortion_model = "plumb_bob"
            info_msg.header.frame_id = camera.getCameraLink().getName()
            info_msg.width = camera.getResolution().width
            info_msg.height = camera.getResolution().height
            info_msg.D = [0.0, 0.0, 0.0, 0.0, 0.0]
            info_msg.K = camera._getCameraIntrinsics()
            info_msg.R = [1, 0, 0, 0, 1, 0, 0, 0, 1]
            info_msg.P = list(info_msg.K)
            info_msg.P.insert(3, 0.0)
            info_msg.P.insert(7, 0.0)
            info_msg.P.append(0.0)

            # Fill the image message
            image_msg = self.image_bridge.cv2_to_imgmsg(frame)
            image_msg.header.frame_id = camera.getCameraLink().getName()

            # Check if the retrieved image is RGB or a depth image            
            if isinstance(camera, CameraDepth):
                image_msg.encoding = "16UC1"
            else:
                image_msg.encoding = "bgr8"

            # Publish the image and the camera info
            image_publisher.publish(image_msg)
            info_publisher.publish(info_msg)

        except AssertionError:
            pass

    def _broadcastJointState(self, joint_state_publisher, extra_joints=None):
        """
        INTERNAL METHOD, publishes the state of the robot's joints into the ROS
        framework

        Parameters:
            joint_state_publisher - The ROS publisher for the JointState
            message, describing the state of the robot's joints
            extra_joints - A dict, describing extra joints to be published. The
            dict should respect the following syntax:
            {"joint_name": joint_value, ...}
        """
        msg_joint_state = JointState()
        msg_joint_state.header = Header()
        msg_joint_state.header.stamp = rospy.get_rostime()
        msg_joint_state.name = list(self.robot.joint_dict)
        msg_joint_state.position = self.robot.getAnglesPosition(
            msg_joint_state.name)

        try:
            assert isinstance(extra_joints, dict)

            for name, value in extra_joints.items():
                msg_joint_state.name += [name]
                msg_joint_state.position += [value]

        except AssertionError:
            pass

        joint_state_publisher.publish(msg_joint_state)

    def _jointAnglesCallback(self, msg):
        """
        INTERNAL METHOD, callback triggered when a message is received on the
        /joint_angles topic

        Parameters:
            msg - a ROS message containing a pose stamped with a speed
            associated to it. The type of the message is the following:
            naoqi_bridge_msgs::JointAnglesWithSpeed. That type can be found in
            the ros naoqi software stack
        """
        joint_list = msg.joint_names
        position_list = list(msg.joint_angles)

        # If the "non official" driver (softbankrobotics-research fork) is
        # used, will try to detect if multiple speeds have been provided. If
        # not, or if the "official" driver is used, the speed attribute of the
        # message will be used
        try:
            assert not OFFICIAL_DRIVER

            if len(msg.speeds) != 0:
                velocity = list(msg.speeds)
            else:
                velocity = msg.speed

        except AssertionError:
            velocity = msg.speed

        self.robot.setAngles(joint_list, position_list, velocity)


class NaoRosWrapper(RosWrapper):
    """
    Class describing a ROS wrapper for the virtual model of Nao, inheriting
    from the RosWrapperClass
    """
    def __init__(self):
        """
        Constructor
        """
        RosWrapper.__init__(self)

    def launchWrapper(self, virtual_nao, ros_namespace, frequency=200):
        """
        Launches the ROS wrapper for the virtual_nao instance

        Parameters:
            virtual_nao - The instance of the simulated model
            ros_namespace - The ROS namespace to be added before the ROS topics
            advertized and subscribed
            frequency - The frequency of the ROS rate that will be used to pace
            the wrapper's main loop
        """
        RosWrapper.launchWrapper(
            self,
            virtual_nao,
            ros_namespace,
            frequency)

    def _initPublishers(self):
        """
        INTERNAL METHOD, initializes the ROS publishers
        """
        self.front_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/image_raw',
            Image,
            queue_size=10)

        self.front_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/camera_info',
            CameraInfo,
            queue_size=10)

        self.bottom_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/image_raw',
            Image,
            queue_size=10)

        self.bottom_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/camera_info',
            CameraInfo,
            queue_size=10)

        self.joint_states_pub = rospy.Publisher(
            '/joint_states',
            JointState,
            queue_size=10)

        self.odom_pub = rospy.Publisher(
            'odom',
            Odometry,
            queue_size=10)

    def _initSubscribers(self):
        """
        INTERNAL METHOD, initializes the ROS subscribers
        """
        rospy.Subscriber(
            '/joint_angles',
            JointAnglesWithSpeed,
            self._jointAnglesCallback)

    def _broadcastCamera(self):
        """
        INTERNAL METHOD, overloading @_broadcastCamera in RosWrapper
        """
        if self.robot.camera_dict[NaoVirtual.ID_CAMERA_TOP].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[NaoVirtual.ID_CAMERA_TOP],
                self.front_cam_pub,
                self.front_info_pub)

        elif self.robot.camera_dict[NaoVirtual.ID_CAMERA_BOTTOM].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[NaoVirtual.ID_CAMERA_BOTTOM],
                self.bottom_cam_pub,
                self.bottom_info_pub)

    def _broadcastJointState(self, joint_state_publisher):
        """
        INTERNAL METHOD, publishes the state of the robot's joints into the ROS
        framework, overloading @_broadcastJointState in RosWrapper

        Parameters:
            joint_state_publisher - The ROS publisher for the JointState
            message, describing the state of the robot's joints (for API
            consistency)
        """
        RosWrapper._broadcastJointState(self, joint_state_publisher)

    def _spin(self):
        """
        INTERNAL METHOD, designed to emulate a ROS spin method
        """
        rate = rospy.Rate(self.frequency)

        try:
            while not self._wrapper_termination:
                rate.sleep()
                self._broadcastJointState(self.joint_states_pub)
                self._broadcastOdometry(self.odom_pub)
                self._broadcastCamera()

        except Exception as e:
            print("Stopping the ROS wrapper: " + str(e))


class RomeoRosWrapper(RosWrapper):
    """
    Class describing a ROS wrapper for the virtual model of Romeo, inheriting
    from the RosWrapperClass
    """
    def __init__(self):
        """
        Constructor
        """
        RosWrapper.__init__(self)

    def launchWrapper(self, virtual_romeo, ros_namespace, frequency=200):
        """
        Launches the ROS wrapper for the virtual_romeo instance

        Parameters:
            virtual_romeo - The instance of the simulated model
            ros_namespace - The ROS namespace to be added before the ROS topics
            advertized and subscribed
            frequency - The frequency of the ROS rate that will be used to pace
            the wrapper's main loop
        """
        RosWrapper.launchWrapper(
            self,
            virtual_romeo,
            ros_namespace,
            frequency)

    def _initPublishers(self):
        """
        INTERNAL METHOD, initializes the ROS publishers
        """
        self.right_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/right/image_raw',
            Image,
            queue_size=10)

        self.right_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/right/camera_info',
            CameraInfo,
            queue_size=10)

        self.left_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/left/image_raw',
            Image,
            queue_size=10)

        self.left_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/left/camera_info',
            CameraInfo,
            queue_size=10)

        self.depth_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/image_raw',
            Image,
            queue_size=10)

        self.depth_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/camera_info',
            CameraInfo,
            queue_size=10)

        self.joint_states_pub = rospy.Publisher(
            '/joint_states',
            JointState,
            queue_size=10)

        self.odom_pub = rospy.Publisher(
            'odom',
            Odometry,
            queue_size=10)

    def _initSubscribers(self):
        """
        INTERNAL METHOD, initializes the ROS subscribers
        """
        rospy.Subscriber(
            '/joint_angles',
            JointAnglesWithSpeed,
            self._jointAnglesCallback)

    def _broadcastCamera(self):
        """
        INTERNAL METHOD, overloading @_broadcastCamera in RosWrapper
        """
        if self.robot.camera_dict[RomeoVirtual.ID_CAMERA_RIGHT].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[RomeoVirtual.ID_CAMERA_RIGHT],
                self.right_cam_pub,
                self.right_info_pub)

        elif self.robot.camera_dict[RomeoVirtual.ID_CAMERA_LEFT].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[RomeoVirtual.ID_CAMERA_LEFT],
                self.left_cam_pub,
                self.left_info_pub)

        elif self.robot.camera_dict[RomeoVirtual.ID_CAMERA_DEPTH].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[RomeoVirtual.ID_CAMERA_DEPTH],
                self.depth_cam_pub,
                self.depth_info_pub)

    def _broadcastJointState(self, joint_state_publisher):
        """
        INTERNAL METHOD, publishes the state of the robot's joints into the ROS
        framework, overloading @_broadcastJointState in RosWrapper

        Parameters:
            joint_state_publisher - The ROS publisher for the JointState
            message, describing the state of the robot's joints (for API
            consistency)
        """
        RosWrapper._broadcastJointState(self, joint_state_publisher)

    def _spin(self):
        """
        INTERNAL METHOD, designed to emulate a ROS spin method
        """
        rate = rospy.Rate(self.frequency)

        try:
            while not self._wrapper_termination:
                rate.sleep()
                self._broadcastJointState(self.joint_states_pub)
                self._broadcastOdometry(self.odom_pub)
                self._broadcastCamera()

        except Exception as e:
            print("Stopping the ROS wrapper: " + str(e))


class PepperRosWrapper(RosWrapper):
    """
    Class describing a ROS wrapper for the virtual model of Pepper, inheriting
    from the RosWrapperClass
    """

    def __init__(self):
        """
        Constructor
        """
        RosWrapper.__init__(self)

    def launchWrapper(self, virtual_pepper, ros_namespace, frequency=200):
        """
        Launches the ROS wrapper for the virtual_pepper instance

        Parameters:
            virtual_pepper - The instance of the simulated model
            ros_namespace - The ROS namespace to be added before the ROS topics
            advertized and subscribed
            frequency - The frequency of the ROS rate that will be used to pace
            the wrapper's main loop
        """
        RosWrapper.launchWrapper(
            self,
            virtual_pepper,
            ros_namespace,
            frequency)

    def _initPublishers(self):
        """
        INTERNAL METHOD, initializes the ROS publishers
        """
        self.front_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/image_raw',
            Image,
            queue_size=10)

        self.front_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/camera_info',
            CameraInfo,
            queue_size=10)

        self.bottom_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/image_raw',
            Image,
            queue_size=10)

        self.bottom_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/camera_info',
            CameraInfo,
            queue_size=10)

        self.depth_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/image_raw',
            Image,
            queue_size=10)

        self.depth_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/camera_info',
            CameraInfo,
            queue_size=10)

        self.laser_pub = rospy.Publisher(
            self.ros_namespace + "/laser",
            LaserScan,
            queue_size=10)

        self.joint_states_pub = rospy.Publisher(
            '/joint_states',
            JointState,
            queue_size=10)

        self.odom_pub = rospy.Publisher(
            '/naoqi_driver/odom',
            Odometry,
            queue_size=10)

    def _initSubscribers(self):
        """
        INTERNAL METHOD, initializes the ROS subscribers
        """
        rospy.Subscriber(
            '/joint_angles',
            JointAnglesWithSpeed,
            self._jointAnglesCallback)

        rospy.Subscriber(
            '/cmd_vel',
            Twist,
            self._velocityCallback)

        rospy.Subscriber(
            '/move_base_simple/goal',
            MovetoPose,
            self._moveToCallback)

        rospy.Subscriber(
            '/move_base_simple/cancel',
            Empty,
            self._killMoveCallback)

    def _broadcastLasers(self, laser_publisher):
        """
        INTERNAL METHOD, publishes the laser values in the ROS framework

        Parameters:
            laser_publisher - The ROS publisher for the LaserScan message,
            corresponding to the laser info of the pepper robot (for API
            consistency)
        """
        if not self.robot.laser_manager.isActive():
            return

        scan = LaserScan()
        scan.header.stamp = rospy.get_rostime()
        scan.header.frame_id = "base_footprint"
        # -120 degres, 120 degres
        scan.angle_min = -2.0944
        scan.angle_max = 2.0944

        # 240 degres FoV, 61 points (blind zones inc)
        scan.angle_increment = (2 * 2.0944) / (15.0 + 15.0 + 15.0 + 8.0 + 8.0)

        # Detection ranges for the lasers in meters, 0.1 to 3.0 meters
        scan.range_min = 0.1
        scan.range_max = 3.0

        # Fill the lasers information
        right_scan = self.robot.getRightLaserValue()
        front_scan = self.robot.getFrontLaserValue()
        left_scan = self.robot.getLeftLaserValue()

        if isinstance(right_scan, list):
            scan.ranges.extend(list(reversed(right_scan)))
            scan.ranges.extend([-1]*8)
        if isinstance(front_scan, list):
            scan.ranges.extend(list(reversed(front_scan)))
            scan.ranges.extend([-1]*8)
        if isinstance(left_scan, list):
            scan.ranges.extend(list(reversed(left_scan)))

        laser_publisher.publish(scan)

    def _broadcastCamera(self):
        """
        INTERNAL METHOD, overloading @_broadcastCamera in RosWrapper
        """
        if self.robot.camera_dict[PepperVirtual.ID_CAMERA_TOP].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[PepperVirtual.ID_CAMERA_TOP],
                self.front_cam_pub,
                self.front_info_pub)

        elif self.robot.camera_dict[PepperVirtual.ID_CAMERA_BOTTOM].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[PepperVirtual.ID_CAMERA_BOTTOM],
                self.bottom_cam_pub,
                self.bottom_info_pub)

        elif self.robot.camera_dict[PepperVirtual.ID_CAMERA_DEPTH].isActive():
            RosWrapper._broadcastCamera(
                self,
                self.robot.camera_dict[PepperVirtual.ID_CAMERA_DEPTH],
                self.depth_cam_pub,
                self.depth_info_pub)

    def _broadcastJointState(self, joint_state_publisher):
        """
        INTERNAL METHOD, publishes the state of the robot's joints into the ROS
        framework, overloading @_broadcastJointState in RosWrapper

        Parameters:
            joint_state_publisher - The ROS publisher for the JointState
            message, describing the state of the robot's joints (for API
            consistency)
        """
        RosWrapper._broadcastJointState(
            self,
            joint_state_publisher,
            extra_joints={"WheelFL": 0.0, "WheelFR": 0.0, "WheelB": 0.0})

    def _velocityCallback(self, msg):
        """
        INTERNAL METHOD, callback triggered when a message is received on the
        /cmd_vel topic

        Parameters:
            msg - a ROS message containing a Twist command
        """
        self.robot.move(msg.linear.x, msg.linear.y, msg.angular.z)

    def _moveToCallback(self, msg):
        """
        INTERNAL METHOD, callback triggered when a message is received on the
        '/move_base_simple/goal' topic. It allows to move the robot's base

        Parameters:
            msg - a ROS message containing a pose stamped with a speed, or a
            simple pose stamped (depending on which version of the naoqi_driver
            is used, the "official" one from ros-naoqi or the "non official"
            softbankrobotics-research fork). The type of the message is the
            following: geometry_msgs::PoseStamped for the "official",
            naoqi_bridge_msgs::PoseStampedWithSpeed for the "non-official".
            An alias is given to the message type: MovetoPose
        """

        if OFFICIAL_DRIVER:
            pose = msg.pose
            frame = 0
            frame_id = msg.header.frame_id
            speed = None
        else:
            pose = msg.pose_stamped.pose
            frame = msg.referenceFrame
            frame_id = msg.pose_stamped.header.frame_id
            speed = msg.speed_percentage *\
                PepperBaseController.MAX_LINEAR_VELOCITY +\
                PepperBaseController.MIN_LINEAR_VELOCITY

        try:
            assert frame not in [
                PepperVirtual.FRAME_ROBOT,
                PepperVirtual.FRAME_WORLD]

            if frame_id == "odom":
                frame = PepperVirtual.FRAME_WORLD
            elif frame_id == "base_footprint":
                frame = PepperVirtual.FRAME_ROBOT
            else:
                raise pybullet.error(
                    "Incorrect reference frame for move_base_simple, please "
                    "modify the content of your message")

        except AssertionError:
            pass

        x = pose.position.x
        y = pose.position.y
        theta = pybullet.getEulerFromQuaternion([
            pose.orientation.x,
            pose.orientation.y,
            pose.orientation.z,
            pose.orientation.w])[-1]

        self.robot.moveTo(
            x,
            y,
            theta,
            frame=frame,
            speed=speed,
            _async=True)

    def _killMoveCallback(self, msg):
        """
        INTERNAL METHOD, callback triggered when a message is received on the
        '/move_base_simple/cancel' topic. This callback is used to stop the
        robot's base from moving

        Parameters:
            msg - an empty ROS message, with the Empty type
        """
        self.robot.moveTo(0, 0, 0, _async=True)

    def _spin(self):
        """
        INTERNAL METHOD, designed to emulate a ROS spin method
        """
        rate = rospy.Rate(self.frequency)

        try:
            while not self._wrapper_termination:
                rate.sleep()
                self._broadcastJointState(self.joint_states_pub)
                self._broadcastOdometry(self.odom_pub)
                self._broadcastLasers(self.laser_pub)
                self._broadcastCamera()

        except Exception as e:
            print("Stopping the ROS wrapper: " + str(e))



#!/usr/bin/env python
# coding: utf-8

from qibullet.robot_module import RobotModule


class Controller(RobotModule):
    """
    Abstract class describing a virtual controller
    """
    def __init__(self, robot_model, physics_client):
        """
        Constructor

        Parameters:
            robot_model - The pybullet model of the robot
            physics_client - The if of the simulated instance in which the
            robot is spawned
        """
        RobotModule.__init__(self, robot_model, physics_client)



#!/usr/bin/env python
# coding: utf-8

import time
import atexit
import weakref
import pybullet
import threading

from qibullet.tools import *
from qibullet.controller import Controller


class BaseController(Controller):
    """
    Class describing a robot base controller
    """
    # _instances = set()
    FRAME_WORLD = 1
    FRAME_ROBOT = 2

    def __init__(self, robot_model, physicsClientId=0):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            physicsClientId - The id of the simulated instance in which the
            robot will be controlled
        """
        Controller.__init__(self, robot_model, physicsClientId)
        self.linear_velocity = 0
        self.angular_velocity = 0
        self.linear_acceleration = 0
        self.angular_acceleration = 0
        self.frame = BaseController.FRAME_ROBOT
        self.pose_init = {}
        self.pose_goal = {}

    def _setGoal(self, x, y, theta, frame):
        """
        INTERNAL METHOD, set the position of the goal to a specific frame.

        Parameters:
            x - position of the goal on the x axis, in meters
            y - position of the goal on the y axis, in meters
            theta - orientation of the goal around the z axis, in radians
            frame - The frame in which the goal is expressed: FRAME_WORLD = 1,
            FRAME_ROBOT = 2
        """
        self.goal = [x, y, theta]
        self.frame = frame

    def _updateGoal(self):
        """
        INTERNAL METHOD, update the position of the goal.
        """
        # get actual position in frame world
        actual_pos, actual_orn = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)

        x, y, theta = self.goal
        # pose x, y, z
        pose_requested = [x, y, 0]

        # orientation requested (euler)
        orn_requested = [0, 0, theta]
        # if we are in frame robot express the position in the frame world
        if self.frame == BaseController.FRAME_ROBOT:
            orn_euler = pybullet.getEulerFromQuaternion(actual_orn)
            pose_requested = [
                pose_requested[0] * math.cos(orn_euler[2])
                - pose_requested[1] * math.sin(orn_euler[2])
                + actual_pos[0],
                pose_requested[0] * math.sin(orn_euler[2])
                + pose_requested[1] * math.cos(orn_euler[2])
                + actual_pos[1],
                0]
            orn_requested = [
                orn_euler[0],
                orn_euler[1],
                orn_euler[2] + theta]
        self.pose_goal["position"] = pose_requested
        self.pose_goal["orientation"] = orn_requested

    def setLinearVelocity(self, linear_velocity):
        """
        Set the linear velocity.

        Parameter:
            linear_velocity : The linear velocity value in m/s
        """
        self.linear_velocity = linear_velocity

    def _setAngularVelocity(self, angular_velocity):
        """
        INTERNAL METHOD, set the angular velocity.

        Parameter:
            angular_velocity : The angular velocity value in rad/s
        """
        self.angular_velocity = angular_velocity

    def _setLinearAcceleration(self, linear_acceleration):
        """
        INTERNAL METHOD, set the linear acceleration.

        Parameter:
            linear_acceleration : The linear acceleration value in m/s^2
        """
        self.linear_acceleration = linear_acceleration

    def _setAngularAcceleration(self, angular_acceleration):
        """
        INTERNAL METHOD, set the angular acceleration.

        Parameter:
            angular_acceleration : The angular acceleration value in rad/s^2
        """
        self.angular_acceleration = angular_acceleration


class PepperBaseController(BaseController):
    """
    Class describing a Pepper base controller
    """
    MAX_LINEAR_VELOCITY = 0.55
    MIN_LINEAR_VELOCITY = 0.1
    MAX_ANGULAR_VELOCITY = 2.0
    MIN_ANGULAR_VELOCITY = 0.3
    MAX_LINEAR_ACCELERATION = 0.55
    MIN_LINEAR_ACCELERATION = 0.1
    MAX_ANGULAR_ACCELERATION = 3.0
    MIN_ANGULAR_ACCELERATION = 0.1

    def __init__(
            self,
            robot_model,
            speed,
            acceleration,
            motion_constraint,
            physicsClientId=0):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            speed - list containing the linear velocity and the angular
            velocity values, in m/s
            acceleration - list containing the linear acceleration and angular
            acceleration values, in m/s^2
            motion_constraint - the pybullet motion constraint applied on the
            robot
            physicsClientId - The id of the simulated instance in which Pepper
            will be controlled
        """
        BaseController.__init__(
            self,
            robot_model,
            physicsClientId=physicsClientId)

        # Set the different speeds and accelerations
        self.setLinearVelocity(speed[0])
        self._setAngularVelocity(speed[1])
        self._setLinearAcceleration(acceleration[0])
        self._setAngularAcceleration(acceleration[1])

        # force applied in the movement
        self.force = 100

        # The robot will stop the movement with a precisio of 0.01 m and 0.02
        # rads
        self.linear_threshold = 0.01
        self.angular_threshold = 0.02
        self.motion_constraint = motion_constraint

    def setLinearVelocity(self, linear_velocity):
        """
        Set the linear velocity.

        Parameter:
            linear_velocity : The linear velocity value in m/s
        """
        if linear_velocity > PepperBaseController.MAX_LINEAR_VELOCITY:
            linear_velocity = PepperBaseController.MAX_LINEAR_VELOCITY

        elif linear_velocity < PepperBaseController.MIN_LINEAR_VELOCITY:
            linear_velocity = PepperBaseController.MIN_LINEAR_VELOCITY

        BaseController.setLinearVelocity(self, linear_velocity)

    def _setAngularVelocity(self, angular_velocity):
        """
        INTERNAL METHOD, set the angular velocity.

        Parameter:
            angular_velocity : The angular velocity value in rad/s
        """
        if angular_velocity > PepperBaseController.MAX_ANGULAR_VELOCITY:
            angular_velocity = PepperBaseController.MAX_ANGULAR_VELOCITY

        elif angular_velocity < PepperBaseController.MIN_ANGULAR_VELOCITY:
            angular_velocity = PepperBaseController.MIN_ANGULAR_VELOCITY

        BaseController._setAngularVelocity(self, angular_velocity)

    def _setLinearAcceleration(self, linear_acceleration):
        """
        INTERNAL METHOD, set the linear acceleration.

        Parameter:
            linear_acceleration : The linear acceleration value in m/s^2
        """
        if linear_acceleration > PepperBaseController.MAX_LINEAR_ACCELERATION:
            linear_acceleration = PepperBaseController.MAX_LINEAR_ACCELERATION

        elif linear_acceleration <\
                PepperBaseController.MIN_LINEAR_ACCELERATION:
            linear_acceleration = PepperBaseController.MIN_LINEAR_ACCELERATION

        BaseController._setLinearAcceleration(self, linear_acceleration)

    def _setAngularAcceleration(self, angular_acceleration):
        """
        INTERNAL METHOD, set the angular acceleration.

        Parameter:
            angular_acceleration : The angular acceleration value in rad/s^2
        """
        if angular_acceleration >\
                PepperBaseController.MAX_ANGULAR_ACCELERATION:
            angular_acceleration =\
                PepperBaseController.MAX_ANGULAR_ACCELERATION

        elif angular_acceleration <\
                PepperBaseController.MIN_ANGULAR_ACCELERATION:
            angular_acceleration =\
                PepperBaseController.MIN_ANGULAR_ACCELERATION

        BaseController._setAngularAcceleration(self, angular_acceleration)

    def moveTo(self, x, y, theta, frame, _async=False):
        """
        Move the robot in frame world or robot (FRAME_WORLD=1, FRAME_ROBOT=2).
        This method can be called synchonously or asynchronously. In the
        asynchronous mode, the function can be called when it's already
        launched, this will update the goal of the motion.

        Parameters:
            x - position of the goal on the x axis, in meters
            y - position of the goal on the y axis, in meters
            theta - orientation of the goal around the z axis, in radians
            frame - The frame in which the goal is expressed: FRAME_WORLD = 1,
            FRAME_ROBOT = 2
            _async - The method is launched in async mode if True, in synch
            mode if False (False by default)
        """
        self._setGoal(x, y, theta, frame)

        if self.module_process.isAlive():
            if _async is False:
                raise pybullet.error(
                    "Already a moveTo asynchronous. Can't "
                    "launch moveTo synchronous")
            self._initProcess()
        elif _async:
            self.module_process = threading.Thread(target=self._moveToProcess)
            self.module_process.start()
        else:
            self._moveToProcess()

    def move(self, x, y, theta):
        """
        Apply a speed on the robot's base.

        Parameters:
            x - Speed on the x axis, in m/s
            y - Speed on the y axis, in m/s
            theta - Rotational speed around the z axis, in rad/s
        """
        # Kill any previous moveTo process running
        self.moveTo(0, 0, 0, frame=BaseController.FRAME_ROBOT, _async=True)

        # Bound the velocity. The max acceleration is not taken into account
        # here, this is a potential improvment
        if abs(x) > PepperBaseController.MAX_LINEAR_VELOCITY:
            x = PepperBaseController.MAX_LINEAR_VELOCITY * (x/abs(x))
        if abs(y) > PepperBaseController.MAX_LINEAR_VELOCITY:
            y = PepperBaseController.MAX_LINEAR_VELOCITY * (y/abs(y))
        if abs(theta) > PepperBaseController.MAX_ANGULAR_VELOCITY:
            theta = PepperBaseController.MAX_ANGULAR_VELOCITY *\
                (theta/abs(theta))

        actual_pos, actual_orn = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)

        # convert actual_orn into euler
        actual_orn = pybullet.getEulerFromQuaternion(actual_orn)

        linear_world_velocity = [
            x * math.cos(actual_orn[2]) - y * math.sin(actual_orn[2]),
            x * math.sin(actual_orn[2]) + y * math.cos(actual_orn[2]),
            0]

        time.sleep(0.02)
        pybullet.resetBaseVelocity(
            self.robot_model,
            linear_world_velocity,
            [0, 0, theta],
            physicsClientId=self.physics_client)

    def _updateConstraint(self):
        """
        INTERNAL METHOD, update the robot's constraint.
        """
        # Change the constraint to the requested position and orientation
        pybullet.changeConstraint(
            self.motion_constraint,
            self.pose_goal["position"],
            jointChildFrameOrientation=pybullet.getQuaternionFromEuler(
                self.pose_goal["orientation"]),
            maxForce=self.force,
            physicsClientId=self.physics_client)

    def _initProcess(self):
        """
        INTERNAL METHOD, initialize the motion process and all variables
        needed.
        """
        # Get actual position in frame world
        self.pose_init["position"], self.pose_init["orientation"] =\
            pybullet.getBasePositionAndOrientation(
                self.robot_model,
                physicsClientId=self.physics_client)

        # convert pose_init orientation in orn_euler
        self.pose_init["orientation"] = pybullet.getEulerFromQuaternion(
            self.pose_init["orientation"]
        )
        self._updateGoal()
        self._updateConstraint()

        # Compute the ratio distance requested on the total distance
        distance = getDistance(
            self.pose_init["position"],
            self.pose_goal["position"])

        self.p_x = 0
        self.p_y = 0
        self.p_theta = 0

        if distance:
            self.p_x = (
                self.pose_goal["position"][0] -
                self.pose_init["position"][0]) / distance
            self.p_y = (
                self.pose_goal["position"][1] -
                self.pose_init["position"][1]) / distance

        theta_to_do = getOrientation(
            self.pose_init["orientation"],
            self.pose_goal["orientation"])

        if abs(theta_to_do):
            self.p_theta = abs(theta_to_do) / theta_to_do

    def _endProcess(self):
        """
        INTERNAL METHOD, stop the robot movement.
        """
        # Change the constraint to the actual position and orientation in
        # order to stop the robot's motion. The force applied is purposely huge
        # to avoid oscillations.
        actual_pos, actual_orn = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)
        pybullet.changeConstraint(
            self.motion_constraint,
            actual_pos,
            jointChildFrameOrientation=actual_orn,
            maxForce=self.force * 10,
            physicsClientId=self.physics_client)
        pybullet.resetBaseVelocity(
            self.robot_model,
            [0, 0, 0],
            [0, 0, 0],
            physicsClientId=self.physics_client)

    def _moveToProcess(self):
        """
        INTERNAL METHOD, process allowing to move the robot's base.
        """
        self._initProcess()
        # actual_pos = self.pose_init["position"]
        # actual_orn = self.pose_init["orientation"]
        init_pos = self.pose_init["position"]
        init_orn = self.pose_init["orientation"]
        actual_pos = init_pos
        actual_orn = init_orn

        while not self._module_termination:
            translation_distance = getDistance(
                actual_pos,
                self.pose_goal["position"])

            # Modulo the orientation pose goal with conversion in quaternion
            modulo_quater_pose_goal = pybullet.getQuaternionFromEuler(
                self.pose_goal["orientation"])
            # Conversion into euler
            modulo_euler_pose_goal = pybullet.getEulerFromQuaternion(
                modulo_quater_pose_goal)
            rotation_distance = abs(getOrientation(
                actual_orn,
                modulo_euler_pose_goal))

            if translation_distance < self.linear_threshold and\
                    rotation_distance < self.angular_threshold:
                break

            actual_pos, actual_orn = pybullet.getBasePositionAndOrientation(
                self.robot_model,
                physicsClientId=self.physics_client)
            # convert actual_orn into euler
            actual_orn = pybullet.getEulerFromQuaternion(actual_orn)

            linear_vel_x = computeVelocity(
                self.linear_acceleration,
                0.05,
                self.linear_velocity,
                getDistance(actual_pos, init_pos),
                getDistance(actual_pos, self.pose_goal["position"]))

            linear_vel_y = linear_vel_x

            angular_vel = computeVelocity(
                self.angular_acceleration,
                0.05,
                self.angular_velocity,
                abs(getOrientation(
                    init_orn,
                    actual_orn)),
                abs(getOrientation(
                    actual_orn,
                    self.pose_goal["orientation"])))

            # If the robot is on the requested position, we set the velocity to
            # 0.
            if abs(actual_pos[0] - self.pose_goal["position"][0]) <=\
                    self.linear_threshold / 2:
                linear_vel_x = 0

            if abs(actual_pos[1] - self.pose_goal["position"][1]) <=\
                    self.linear_threshold / 2:
                linear_vel_y = 0

            if abs(getOrientation(
                    actual_orn, self.pose_goal["orientation"])) <=\
                    self.angular_threshold:
                angular_vel = 0

            # Reset the velocity of the robot
            time.sleep(0.02)
            pybullet.resetBaseVelocity(
                self.robot_model,
                [linear_vel_x * self.p_x, linear_vel_y * self.p_y, 0],
                [0, 0, angular_vel * self.p_theta],
                physicsClientId=self.physics_client)

        self._endProcess()



#!/usr/bin/env python
# coding: utf-8

import sys
import pybullet
from qibullet.camera import *
from qibullet.link import Link
from qibullet.joint import Joint

IS_VERSION_PYTHON_3 = sys.version_info[0] >= 3


class RobotVirtual:
    """
    Mother class representing a virtual robot
    """

    def __init__(self, description_file):
        """
        Constructor

        Parameters:
            description_file - The file giving the description of the virtual
            robot. For now, only URDF is handled
        """
        self.description_file = description_file
        self.physics_client = 0
        self.camera_dict = dict()
        self.joint_dict = dict()
        self.link_dict = dict()

    def loadRobot(self, translation, quaternion, physicsClientId=0):
        """
        Loads the robot into a simulation, loads the joints and the links
        descriptions. The joints are set to 0 rad.

        Parameters:
            translation - List containing 3 elements, the translation [x, y, z]
            of the robot in the WORLD frame
            quaternion - List containing 4 elements, the quaternion
            [x, y, z, q] of the robot in the WORLD frame
            physicsClientId - The id of the simulated instance in which the
            robot is supposed to be loaded

        Returns:
            boolean - True if the method ran correctly, False otherwise
        """
        try:
            self.physics_client = physicsClientId
            self.robot_model = pybullet.loadURDF(
                self.description_file,
                translation,
                quaternion,
                useFixedBase=False,
                globalScaling=1.0,
                physicsClientId=self.physics_client,
                flags=pybullet.URDF_USE_SELF_COLLISION |
                pybullet.URDF_USE_MATERIAL_COLORS_FROM_MTL)

        except pybullet.error as e:
            raise pybullet.error("Cannot load robot model: " + str(e))

        for i in range(pybullet.getNumJoints(
                self.robot_model,
                physicsClientId=self.physics_client)):
            if IS_VERSION_PYTHON_3:
                # PYTHON 3 version needs a conversion bytes to str
                joint_info = pybullet.getJointInfo(
                    self.robot_model,
                    i,
                    physicsClientId=self.physics_client)
                self.link_dict[joint_info[12].decode('utf-8')] =\
                    Link(joint_info)

                if joint_info[2] == pybullet.JOINT_PRISMATIC or\
                        joint_info[2] == pybullet.JOINT_REVOLUTE:
                    self.joint_dict[joint_info[1].decode('utf-8')] =\
                        Joint(joint_info)
            else:
                # PYTHON 2 Version
                joint_info = pybullet.getJointInfo(
                    self.robot_model,
                    i,
                    physicsClientId=self.physics_client)

                self.link_dict[joint_info[12]] = Link(joint_info)

                if joint_info[2] == pybullet.JOINT_PRISMATIC or\
                        joint_info[2] == pybullet.JOINT_REVOLUTE:
                    self.joint_dict[joint_info[1]] = Joint(joint_info)

    def getRobotModel(self):
        """
        Returns the pybullet model to which the module is associated.

        Returns:
            robot_model - The pybullet model of the robot
        """
        return self.robot_model

    def getPhysicsClientId(self):
        """
        Returns the id of the simulated instance in which the module is loaded.

        Returns:
            physics_client - The id of the simulation in which the robot
            (possessing the module) is spawned
        """
        return self.physics_client

    def setAngles(self, joint_names, joint_values, percentage_speeds):
        """
        Set angles on the robot's joints. Tests have to be performed by the
        child class to guarantee the validity of the input parameters.

        Parameters:
            joint_names - List of string containing the name of the joints
            to be controlled
            joint_values - List of values corresponding to the angles in
            radians to be applied
            percentage_speeds - Percentages of the max speed to be used for
            each joint, has to be strictly superior to 0 and inferior or equal
            to 1
        """
        try:
            assert len(joint_names) ==\
                len(joint_values) ==\
                len(percentage_speeds)

            assert all(
                speed >= 0.0 and speed <= 1.0 for speed in percentage_speeds)

        except AssertionError:
            raise pybullet.error("Error in the setAngles parameters")

        for joint_name, joint_value, percentage_speed in zip(
                joint_names,
                joint_values,
                percentage_speeds):

            joint_speed =\
                self.joint_dict[joint_name].getMaxVelocity() *\
                percentage_speed

            pybullet.setJointMotorControl2(
                self.robot_model,
                self.joint_dict[joint_name].getIndex(),
                pybullet.POSITION_CONTROL,
                targetPosition=joint_value,
                maxVelocity=joint_speed,
                force=self.joint_dict[joint_name].getMaxEffort(),
                physicsClientId=self.physics_client)

    def getAnglesPosition(self, joint_names):
        """
        Gets the position of the robot's joints in radians. If one of the joint
        doesn't exist, the method will raise a KeyError.

        Parameters:
            joint_names - List of string containing the names of the joints

        Returns:
            joint_positions - List of floats containing the joint's positions
        """
        indexes = [self.joint_dict[name].getIndex() for name in joint_names]

        return [state[0] for state in pybullet.getJointStates(
            self.robot_model,
            indexes,
            physicsClientId=self.physics_client)]

    def getAnglesVelocity(self, joint_names):
        """
        Gets the velocity of the robot's joints in rad/s. If one of the joint
        doesn't exist, the method will raise a KeyError.

        Parameters:
            joint_names - List of string containing the names of the joints

        Returns:
            joint_velocities - List of floats containing the joint's velocities
        """
        indexes = [self.joint_dict[name].getIndex() for name in joint_names]

        return [state[1] for state in pybullet.getJointStates(
            self.robot_model,
            indexes,
            physicsClientId=self.physics_client)]

    def subscribeCamera(self, camera_id, resolution=Camera.K_QVGA):
        """
        Subscribe to the camera holding the camera id (for instance,
        PepperVirtual.ID_CAMERA_TOP). This method returns a handle for the
        camera, needed to get frames, the resolution / link of the camera and
        the camera object. Bear in mind that when subscribing, a background
        process is started continuously retrieving and treating the
        corresponding camera frames. The handle is also used to unsubscribe
        from a camera, stopping the background process associated

        Parameters:
            camera_id - The id of the camera to be subscribed
            resolution - CameraResolution object, the resolution of the camera

        Returns:
            handle - The associated camera handle
        """
        try:
            return self.camera_dict[camera_id].subscribe(resolution)

        except KeyError:
            print("This camera does not exist, use a valid camera id")

    def unsubscribeCamera(self, handle):
        """
        Unsubscribe from a camera, using the specified handle

        Parameters:
            handle - The handle retrieved when subscribing to the camera

        Returns:
            success - Boolean, True if unsubscribed successfully, False
            otherwise
        """
        try:
            return Camera._getCameraFromHandle(handle).unsubscribe()

        except KeyError:
            print("Invalid handle, nothing to unsubscribe from")
            return False

    def getCameraFrame(self, handle):
        """
        Returns a frame form the camera associated to the specified handle
        object. Be advised that the subscribeCamera method needs to be called
        beforehand, otherwise a pybullet error will be raised.

        Parameters:
            handle - The handle retrieved when subscribing to the camera

        Returns:
            frame - The current camera frame as a formatted numpy array,
            directly exploitable from OpenCV
        """
        try:
            return Camera._getCameraFromHandle(handle).getFrame()

        except KeyError:
            raise pybullet.error("Invalid handle, cannot retrieve any frame")

    def getCameraResolution(self, handle):
        """
        Returns the resolution of the camera associated to the specified
        handle. Be advised that the subscribeCamera method needs to be called
        beforehand, otherwise a pybullet error will be raised.

        Parameters:
            handle - The handle retrieved when subscribing to the camera

        Returns:
            resolution - a CameraResolution object describing the resolution of
            the robot camera associated to the handle
        """
        try:
            return Camera._getCameraFromHandle(handle).getResolution()

        except KeyError:
            raise pybullet.error("Invalid handle, resolution unavailable")

    def getCameraLink(self, handle):
        """
        Returns the link of the camera associated to the specified handle. Be
        advised that the subscribeCamera method needs to be called beforehand,
        otherwise a pybullet error will be raised.

        Parameters:
            handle - The handle retrieved when subscribing to the camera

        Returns:
            link - a Link object describing the link to which the robot camera
            associated to the handle is attached
        """
        try:
            return Camera._getCameraFromHandle(handle).getCameraLink()

        except KeyError:
            raise pybullet.error("Invalid handle, cannot retrieve any link")

    def getCamera(self, handle):
        """
        Returns the robot camera associated to the specidied handle. Be advised
        that the subscribeCamera method needs to be called beforehand,
        otherwise a pybullet error will be raised.

        Parameters:
            handle - The handle retrieved when subscribing to the camera

        Returns:
            camera - Camera (CameraRgb or CameraDepth) object, the
            robot camera associated to the handle.
        """
        try:
            return Camera._getCameraFromHandle(handle)

        except KeyError:
            raise pybullet.error("Invalid handle, no associated camera")

    def getPosition(self):
        """
        Gets the position of the robot's base in the world frame.

        Returns:
            x - The position of the robot's base on the x axis, in meters
            y - The positions of the robot's base on the y axis in meters
            theta - The rotation of the robot's base on the z axis in meters
        """
        position, quaternions = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)

        theta = pybullet.getEulerFromQuaternion(quaternions)[2]
        return position[0], position[1], theta

    def isSelfColliding(self, link_names):
        """
        Specifies if a link is colliding with the rest of the virtual robot.

        Parameters:
            link_names - String or list of string containing the names of the
            links to be checked for self collision. WARNING: only the links
            with corresponding meshes should be used, otherwise the link cannot
            self collide

        Returns:
            self_colliding - Boolean, if True at least one of the links is self
            colliding
        """
        try:
            if type(link_names) is str:
                assert link_names in self.link_dict.keys()
                names = [link_names]
            else:
                assert set(link_names).issubset(self.link_dict.keys())
                names = list(link_names)

            for name in names:
                contact_tuple = pybullet.getContactPoints(
                    bodyA=self.robot_model,
                    bodyB=self.robot_model,
                    linkIndexA=self.link_dict[name].getIndex(),
                    physicsClientId=self.physics_client)
                contact_tuple += pybullet.getContactPoints(
                    bodyA=self.robot_model,
                    bodyB=self.robot_model,
                    linkIndexB=self.link_dict[name].getIndex(),
                    physicsClientId=self.physics_client)

                if len(contact_tuple) != 0:
                    return True

            return False

        except AssertionError:
            raise pybullet.error(
                "Unauthorized link checking for self collisions")



#!/usr/bin/env python
# coding: utf-8

import os
import pybullet
import qibullet.tools as tools
from qibullet.camera import *
from qibullet.robot_posture import RomeoPosture
from qibullet.robot_virtual import RobotVirtual


class RomeoVirtual(RobotVirtual):
    """
    Class representing the virtual instance of the Romeo robot
    """
    ID_CAMERA_RIGHT = 0
    ID_CAMERA_LEFT = 1
    ID_CAMERA_DEPTH = 2
    FRAME_WORLD = 1
    FRAME_ROBOT = 2
    URDF_FILE = "/romeo.urdf"
    P_STAND = RomeoPosture("Stand")
    P_STAND_INIT = RomeoPosture("StandInit")
    P_STAND_ZERO = RomeoPosture("StandZero")
    P_CROUCH = RomeoPosture("Crouch")

    def __init__(self):
        """
        Constructor
        """
        # Install the robot meshes and URDFs if they are not already installed.
        # The installation process won't be covered by the unit tests
        if not tools._check_resources_installed():  # pragma: no cover
            tools._install_resources()

        # Specify the URDF path
        RobotVirtual.__init__(
            self,
            tools._get_resources_folder() + RomeoVirtual.URDF_FILE)

        self.camera_right = None
        self.camera_left = None
        self.camera_depth = None

    def loadRobot(self, translation, quaternion, physicsClientId=0):
        """
        Overloads @loadRobot from the @RobotVirtual class, loads the robot into
        the simulated instance. This method also updates the max velocity of
        the robot's fingers, adds self collision filters to the model and
        defines the cameras of the model in the camera_dict.

        Parameters:
            translation - List containing 3 elements, the translation [x, y, z]
            of the robot in the WORLD frame
            quaternion - List containing 4 elements, the quaternion
            [x, y, z, q] of the robot in the WORLD frame
            physicsClientId - The id of the simulated instance in which the
            robot is supposed to be loaded

        Returns:
            boolean - True if the method ran correctly, False otherwise
        """
        pybullet.setAdditionalSearchPath(
            os.path.dirname(os.path.realpath(__file__)),
            physicsClientId=physicsClientId)

        # Add 0.50 meters on the z component, avoing to spawn NAO in the ground
        translation = [translation[0], translation[1], translation[2] + 1.05]

        RobotVirtual.loadRobot(
            self,
            translation,
            quaternion,
            physicsClientId=physicsClientId)

        balance_constraint = pybullet.createConstraint(
            parentBodyUniqueId=self.robot_model,
            parentLinkIndex=-1,
            childBodyUniqueId=-1,
            childLinkIndex=-1,
            jointType=pybullet.JOINT_FIXED,
            jointAxis=[0, 0, 0],
            parentFramePosition=[0, 0, 0],
            parentFrameOrientation=[0, 0, 0, 1],
            childFramePosition=translation,
            childFrameOrientation=quaternion,
            physicsClientId=self.physics_client)

        self.goToPosture("Stand", 1.0)

        pybullet.setCollisionFilterPair(
            self.robot_model,
            self.robot_model,
            self.link_dict["REye"].getIndex(),
            self.link_dict["LEye"].getIndex(),
            0,
            physicsClientId=self.physics_client)

        for link in ["torso", "HeadRoll_link"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict["NeckPitch_link"].getIndex(),
                self.link_dict[link].getIndex(),
                0,
                physicsClientId=self.physics_client)

        for side in ["R", "L"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side + "Eye"].getIndex(),
                self.link_dict["HeadRoll_link"].getIndex(),
                0,
                physicsClientId=self.physics_client)
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side + "Thigh"].getIndex(),
                self.link_dict["body"].getIndex(),
                0,
                physicsClientId=self.physics_client)
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side.lower() + "_ankle"].getIndex(),
                self.link_dict[side + "Tibia"].getIndex(),
                0,
                physicsClientId=self.physics_client)
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side + "ShoulderYaw_link"].getIndex(),
                self.link_dict["torso"].getIndex(),
                0,
                physicsClientId=self.physics_client)
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side + "WristRoll_link"].getIndex(),
                self.link_dict[side.lower() + "_wrist"].getIndex(),
                0,
                physicsClientId=self.physics_client)

            for link in ["ShoulderYaw_link", "WristYaw_link"]:
                pybullet.setCollisionFilterPair(
                    self.robot_model,
                    self.robot_model,
                    self.link_dict[side + link].getIndex(),
                    self.link_dict[side + "Elbow"].getIndex(),
                    0,
                    physicsClientId=self.physics_client)

            for name, link in self.link_dict.items():
                if side + "Finger" in name or\
                   side + "Thumb" in name:
                    pybullet.setCollisionFilterPair(
                        self.robot_model,
                        self.robot_model,
                        self.link_dict[side.lower() + "_wrist"].getIndex(),
                        link.getIndex(),
                        0,
                        physicsClientId=self.physics_client)

        for joint in self.joint_dict.values():
            pybullet.resetJointState(
                self.robot_model,
                joint.getIndex(),
                0.0)

        pybullet.removeConstraint(
            balance_constraint,
            physicsClientId=self.physics_client)

        for joint_name in list(self.joint_dict):
            if 'RFinger' in joint_name or 'RThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["RHand"].getMaxVelocity())
            elif 'LFinger' in joint_name or 'LThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["LHand"].getMaxVelocity())

        camera_right = CameraRgb(
            self.robot_model,
            RomeoVirtual.ID_CAMERA_RIGHT,
            self.link_dict["CameraRightEye_optical_frame"],
            hfov=60.9,
            vfov=47.6,
            physicsClientId=self.physics_client)

        camera_left = CameraRgb(
            self.robot_model,
            RomeoVirtual.ID_CAMERA_LEFT,
            self.link_dict["CameraLeftEye_optical_frame"],
            hfov=60.9,
            vfov=47.6,
            physicsClientId=self.physics_client)

        camera_depth = CameraDepth(
            self.robot_model,
            RomeoVirtual.ID_CAMERA_DEPTH,
            self.link_dict["CameraDepth_optical_frame"],
            hfov=58.0,
            vfov=45.0,
            physicsClientId=self.physics_client)

        self.camera_dict = {
            RomeoVirtual.ID_CAMERA_RIGHT: camera_right,
            RomeoVirtual.ID_CAMERA_LEFT: camera_left,
            RomeoVirtual.ID_CAMERA_DEPTH: camera_depth}

    def setAngles(self, joint_names, joint_values, percentage_speed):
        """
        Overloads @setAngles from the @RobotVirtual class. Handles the finger
        mimic behaviour.

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints to be controlled
            joint_values - List of values (or value if only one joint)
            corresponding to the angles in radians to be applied
            percentage_speed - Percentage (or percentages) of the max speed to
            be used for the movement
        """
        try:
            if type(joint_names) is str:
                assert type(joint_values) is int or type(joint_values) is float
                names = [joint_names]
                values = [joint_values]
            else:
                assert type(joint_names) is type(joint_values) is list
                names = list(joint_names)
                values = list(joint_values)

            if isinstance(percentage_speed, list):
                speeds = list(percentage_speed)
            else:
                speeds = [percentage_speed]*len(names)

        except AssertionError:
            raise pybullet.error("Error in the parameters given to the\
                setAngles method")

        for hand in ["RHand", "LHand"]:
            for i in range(names.count(hand)):
                index = names.index(hand)
                value = values[index]
                speed = speeds[index]
                names.pop(index)
                values.pop(index)
                speeds.pop(index)
                finger_names, finger_values = self._mimicHand(hand, value)
                names.extend(finger_names)
                values.extend(finger_values)
                speeds.extend([speed]*len(finger_names))

        RobotVirtual.setAngles(
            self,
            names,
            values,
            speeds)

    def getAnglesPosition(self, joint_names):
        """
        Overloads @getAnglesPosition from the @RobotVirtual class. Handles the
        finger mimicked position for the hands (when getting the position of
        RHand or LHand, will return the hand's opening percentage).

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints

        Returns:
            joint_positions - List of float (or float if only one joint)
            containing the joint's positions in radians
        """
        if type(joint_names) is str:
            names = [joint_names]
        else:
            names = list(joint_names)

        joint_positions = RobotVirtual.getAnglesPosition(self, names)

        for hand, finger in zip(
                ["RHand", "LHand"],
                ["RFinger11", "LFinger11"]):
            for i in range(names.count(hand)):
                index = names.index(hand)
                joint_positions[index] =\
                    RobotVirtual.getAnglesPosition(self, [finger]).pop() /\
                    (1/(self.joint_dict[finger].getUpperLimit() -
                        self.joint_dict[finger].getLowerLimit())) +\
                    self.joint_dict[finger].getLowerLimit()

        if len(joint_positions) == 1:
            return joint_positions.pop()
        else:
            return joint_positions

    def getAnglesVelocity(self, joint_names):
        """
        Overloads @getAnglesVelocity from the @RobotVirtual class. The method
        won't return the velocity of RHand and LHand joints.

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints

        Returns:
            joint_velocities - List of float (or float if only one joint)
            containing the joint's velocities in rad/s
        """
        if type(joint_names) is str:
            names = [joint_names]
        else:
            names = list(joint_names)

        joint_velocities = RobotVirtual.getAnglesVelocity(self, names)

        if len(joint_velocities) == 1:
            return joint_velocities.pop()
        else:
            return joint_velocities

    def goToPosture(self, posture_name, percentage_speed):
        """
        Position the virtual robot into a particular posture. The different
        available postures are NaoPosture objects.

        Parameters:
            posture_name - String containing the name of the posture. The
            posture name is not case-sensitive
            percentage_speed - Percentage of the max speed to be used for the
            movement

        Returns:
            Boolean - True if the posture can be applied, False otherwise
        """
        posture_list = [
            RomeoVirtual.P_STAND,
            RomeoVirtual.P_STAND_INIT,
            RomeoVirtual.P_STAND_ZERO,
            RomeoVirtual.P_CROUCH]

        for posture in posture_list:
            if posture.isPostureName(posture_name):
                self.setAngles(
                    posture.getPostureJointNames(),
                    posture.getPostureJointValues(),
                    percentage_speed)

                return True

        return False

    def _mimicHand(
            self,
            hand,
            value,
            multiplier=0.988205,
            thumb_multiplier=1.74533,
            offset=0.0):
        """
        Used to propagate a joint value on the fingers attached to the hand.
        The formula used to mimic a joint is the following:

        finger_value = (hand_value * multiplier) + offset

        Parameters:
            hand - String, RHand or LHand
            value - The joint value to be propagated
            multiplier - The multiplier coefficient for the fingers and the
            sections 2 and 3 of the thumbs (0.988205 by default)
            thumb_multiplier - The multiplier coefficient for the first section
            of the thumb (1.74533 by default)
            offset - The offset coefficient (0.0 by default)

        Returns:
            finger_names - Names of the finger to be controlled
            finger_values - Values of the fingers to be controlled
        """
        finger_names = list()
        finger_values = list()

        for joint_name in self.joint_dict.keys():
            if hand[0] + "Thumb1" in joint_name:
                finger_names.append(joint_name)
                finger_values.append((value * thumb_multiplier) + offset)
            elif (hand[0] + "Finger") in joint_name or\
                 (hand[0] + "Thumb") in joint_name:
                finger_names.append(joint_name)
                finger_values.append((value * multiplier) + offset)

        return finger_names, finger_values



#!/usr/bin/env python
# coding: utf-8

import sys
IS_VERSION_PYTHON_3 = sys.version_info[0] >= 3


class Joint:
    """
    Class describing a robot joint
    """

    def __init__(self, joint_info):
        """
        Constructor

        Parameters:
            name - The name of the joint
            joint_info - Informations returned by the getJointInfo API
        """
        self.index = joint_info[0]
        self.name = joint_info[1]
        self.lower_limit = joint_info[8]
        self.upper_limit = joint_info[9]
        self.max_effort = joint_info[10]
        self.max_velocity = joint_info[11]

        if IS_VERSION_PYTHON_3:
            self.name = self.name.decode('utf-8')

    def __eq__(self, joint):
        """
        Overloading the equal operator

        Parameters:
            joint - the comparing joint
        """
        try:
            assert self.index == joint.index
            assert self.name == joint.name
            assert self.lower_limit == joint.lower_limit
            assert self.upper_limit == joint.upper_limit
            assert self.max_effort == joint.max_effort
            assert self.max_velocity == joint.max_velocity
            return True

        except AssertionError:
            return False

    def getIndex(self):
        """
        Getter for the index parameter

        Returns:
            index - The joint's index
        """
        return self.index

    def getName(self):
        """
        Getter for the name parameter

        Returns:
            name - The joint's name
        """
        return self.name

    def getLowerLimit(self):
        """
        Getter for the lower limit parameter

        Returns:
            lower_limit - The lower limit of the joint in rad
        """
        return self.lower_limit

    def getUpperLimit(self):
        """
        Getter for the upper limit parameter

        Returns:
            upper_limit - The upper limit of the joint in rad
        """
        return self.upper_limit

    def getMaxEffort(self):
        """
        Getter for the max effort parameter

        Returns:
            max_effort - The max effort for the joint in N.m
        """
        return self.max_effort

    def setMaxEffort(self, max_effort):
        """
        Setter for the max effort parameter

        Parameters:
            max_effort - The max effort for the joint in N.m
        """
        self.max_effort = max_effort

    def getMaxVelocity(self):
        """
        Getter for the max velocity parameter

        Returns:
            max_velocity - The max velocity of the joint in rad/s
        """
        return self.max_velocity

    def setMaxVelocity(self, max_velocity):
        """
        Setter for the max velocity parameter

        Parameters:
            max_velocity - The max velocity for the current joint in rad/s
        """
        self.max_velocity = max_velocity



#!/usr/bin/env python
# coding: utf-8

import os
import pybullet
import qibullet.tools as tools
from qibullet.laser import *
from qibullet.camera import *
from qibullet.base_controller import *
from qibullet.robot_posture import PepperPosture
from qibullet.robot_virtual import RobotVirtual


class PepperVirtual(RobotVirtual):
    """
    Class representing the virtual instance of the Pepper robot
    """
    ID_CAMERA_TOP = 0
    ID_CAMERA_BOTTOM = 1
    ID_CAMERA_DEPTH = 2
    FRAME_WORLD = 1
    FRAME_ROBOT = 2
    URDF_FILE = "/pepper.urdf"
    P_STAND = PepperPosture("Stand")
    P_STAND_INIT = PepperPosture("StandInit")
    P_STAND_ZERO = PepperPosture("StandZero")
    P_CROUCH = PepperPosture("Crouch")

    def __init__(self):
        """
        Constructor
        """
        # Install the robot meshes and URDFs if they are not already installed.
        # The installation process won't be covered by the unit tests
        if not tools._check_resources_installed():  # pragma: no cover
            tools._install_resources()

        # Specify the URDF path
        RobotVirtual.__init__(
            self,
            tools._get_resources_folder() + PepperVirtual.URDF_FILE)

        self.motion_constraint = None
        # Default speed (in m/s) xy : 0.35, min : 0.1, max : 0.55
        self.linear_velocity = 0.35
        # Default speed (in rad/s) theta : 1.0, min : 0.3, max : 2.0
        self.angular_velocity = 1.0
        # Default acc (in m/s^2 xy : 0.3, min : 0.1, max : 0.55
        self.linear_acceleration = 0.3
        # Default acc (in rad/s^2 theta : 0.75, min : 0.1, max : 3.0
        self.angular_acceleration = 0.3

    def loadRobot(self, translation, quaternion, physicsClientId=0):
        """
        Overloads @loadRobot from the @RobotVirtual class, loads the robot into
        the simulated instance. This method also updates the max velocity of
        the robot's fingers, adds self collision filters to the model, adds a
        motion constraint to the model and defines the cameras of the model in
        the camera_dict.

        Parameters:
            translation - List containing 3 elements, the translation [x, y, z]
            of the robot in the WORLD frame
            quaternion - List containing 4 elements, the quaternion
            [x, y, z, q] of the robot in the WORLD frame
            physicsClientId - The id of the simulated instance in which the
            robot is supposed to be loaded

        Returns:
            boolean - True if the method ran correctly, False otherwise
        """
        pybullet.setAdditionalSearchPath(
            os.path.dirname(os.path.realpath(__file__)),
            physicsClientId=physicsClientId)

        RobotVirtual.loadRobot(
            self,
            translation,
            quaternion,
            physicsClientId=physicsClientId)

        for link in ["Hip", "Pelvis", "Head"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict["torso"].getIndex(),
                self.link_dict[link].getIndex(),
                0,
                physicsClientId=self.physics_client)

        for shoulder_roll_link in ["RBicep", "LBicep"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict["torso"].getIndex(),
                self.link_dict[shoulder_roll_link].getIndex(),
                0,
                physicsClientId=self.physics_client)

        for side in ["R", "L"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict[side + "ForeArm"].getIndex(),
                self.link_dict[side + "Bicep"].getIndex(),
                0,
                physicsClientId=self.physics_client)

        for name, link in self.link_dict.items():
            for wrist in ["r_wrist", "l_wrist"]:
                if wrist[0] + "finger" in name.lower() or\
                   wrist[0] + "thumb" in name.lower():
                    pybullet.setCollisionFilterPair(
                        self.robot_model,
                        self.robot_model,
                        self.link_dict[wrist].getIndex(),
                        link.getIndex(),
                        0,
                        physicsClientId=self.physics_client)

        for joint_name in list(self.joint_dict):
            if 'RFinger' in joint_name or 'RThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["RHand"].getMaxVelocity())
            elif 'LFinger' in joint_name or 'LThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["LHand"].getMaxVelocity())
            elif "Wheel" in joint_name:
                self.joint_dict.pop(joint_name)

        camera_top = CameraRgb(
            self.robot_model,
            PepperVirtual.ID_CAMERA_TOP,
            self.link_dict["CameraTop_optical_frame"],
            hfov=56.3,
            vfov=43.7,
            physicsClientId=self.physics_client)

        camera_bottom = CameraRgb(
            self.robot_model,
            PepperVirtual.ID_CAMERA_BOTTOM,
            self.link_dict["CameraBottom_optical_frame"],
            hfov=56.3,
            vfov=43.7,
            physicsClientId=self.physics_client)

        camera_depth = CameraDepth(
            self.robot_model,
            PepperVirtual.ID_CAMERA_DEPTH,
            self.link_dict["CameraDepth_optical_frame"],
            hfov=58.0,
            vfov=45.0,
            physicsClientId=self.physics_client)

        self.camera_dict = {
            PepperVirtual.ID_CAMERA_TOP: camera_top,
            PepperVirtual.ID_CAMERA_BOTTOM: camera_bottom,
            PepperVirtual.ID_CAMERA_DEPTH: camera_depth}

        self.motion_constraint = pybullet.createConstraint(
            parentBodyUniqueId=self.robot_model,
            parentLinkIndex=-1,
            childBodyUniqueId=-1,
            childLinkIndex=-1,
            jointType=pybullet.JOINT_FIXED,
            jointAxis=[0, 0, 0],
            parentFramePosition=[0, 0, 0],
            parentFrameOrientation=[0, 0, 0, 1],
            childFramePosition=[translation[0], translation[1], 0],
            childFrameOrientation=quaternion,
            physicsClientId=self.physics_client)

        self.laser_manager = Laser(
            self.robot_model,
            self.link_dict["Tibia"].getIndex(),
            physicsClientId=self.physics_client)

        self.base_controller = PepperBaseController(
            self.robot_model,
            [self.linear_velocity, self.angular_velocity],
            [self.linear_acceleration, self.angular_acceleration],
            self.motion_constraint,
            physicsClientId=self.physics_client)

        self.goToPosture("StandZero", 1.0)

    def moveTo(self, x, y, theta, frame=FRAME_ROBOT, speed=None, _async=False):
        """
        Move the robot in frame world or robot (FRAME_WORLD=1, FRAME_ROBOT=2).
        This method can be called synchonously or asynchronously. In the
        asynchronous mode, the function can be called when it's already
        launched, this will update the goal of the motion.

        Parameters:
            x - position of the goal on the x axis, in meters
            y - position of the goal on the y axis, in meters
            theta - orientation of the goal around the z axis, in radians
            frame - The frame in which the goal is expressed: FRAME_WORLD = 1,
            FRAME_ROBOT = 2
            speed - The desired linear velocity, in m/s
            _async - The method is launched in async mode if True, in synch
            mode if False (False by default)
        """
        if speed is not None:
            self.base_controller.setLinearVelocity(speed)

        self.base_controller.moveTo(x, y, theta, frame, _async=_async)

    def move(self, x, y, theta):
        """
        Apply a speed on the robot's base.

        Parameters:
            x - Speed on the x axis, in m/s
            y - Speed on the y axis, in m/s
            theta - Rotational speed around the z axis, in rad/s
        """
        self.base_controller.move(x, y, theta)

    def setAngles(self, joint_names, joint_values, percentage_speed):
        """
        Overloads @setAngles from the @RobotVirtual class. Handles the finger
        mimic behaviour.

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints to be controlled
            joint_values - List of values (or value if only one joint)
            corresponding to the angles in radians to be applied
            percentage_speed - Percentage (or percentages) of the max speed to
            be used for the movement
        """
        try:
            if type(joint_names) is str:
                assert type(joint_values) is int or type(joint_values) is float
                names = [joint_names]
                values = [joint_values]
            else:
                assert type(joint_names) is type(joint_values) is list
                names = list(joint_names)
                values = list(joint_values)

            if isinstance(percentage_speed, list):
                speeds = list(percentage_speed)
            else:
                speeds = [percentage_speed]*len(names)

        except AssertionError:
            raise pybullet.error("Error in the parameters given to the\
                setAngles method")

        for hand in ["RHand", "LHand"]:
            for i in range(names.count(hand)):
                index = names.index(hand)
                value = values[index]
                speed = speeds[index]
                names.pop(index)
                values.pop(index)
                speeds.pop(index)
                finger_names, finger_values = self._mimicHand(hand, value)
                names.extend(finger_names)
                values.extend(finger_values)
                speeds.extend([speed]*len(finger_names))

        RobotVirtual.setAngles(
            self,
            names,
            values,
            speeds)

    def getAnglesPosition(self, joint_names):
        """
        Overloads @getAnglesPosition from the @RobotVirtual class. Handles the
        finger mimicked position for the hands (when getting the position of
        RHand or LHand, will return the hand's opening percentage).

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints

        Returns:
            joint_positions - List of float (or float if only one joint)
            containing the joint's positions in radians
        """
        if type(joint_names) is str:
            names = [joint_names]
        else:
            names = list(joint_names)

        joint_positions = RobotVirtual.getAnglesPosition(self, names)

        for hand, finger in zip(
                ["RHand", "LHand"],
                ["RFinger11", "LFinger11"]):
            for i in range(names.count(hand)):
                index = names.index(hand)
                joint_positions[index] =\
                    RobotVirtual.getAnglesPosition(self, [finger]).pop() /\
                    (1/(self.joint_dict[finger].getUpperLimit() -
                        self.joint_dict[finger].getLowerLimit())) +\
                    self.joint_dict[finger].getLowerLimit()

        if len(joint_positions) == 1:
            return joint_positions.pop()
        else:
            return joint_positions

    def getAnglesVelocity(self, joint_names):
        """
        Overloads @getAnglesVelocity from the @RobotVirtual class. The method
        won't return the velocity of RHand and LHand joints.

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints

        Returns:
            joint_velocities - List of float (or float if only one joint)
            containing the joint's velocities in rad/s
        """
        if type(joint_names) is str:
            names = [joint_names]
        else:
            names = list(joint_names)

        joint_velocities = RobotVirtual.getAnglesVelocity(self, names)

        if len(joint_velocities) == 1:
            return joint_velocities.pop()
        else:
            return joint_velocities

    def goToPosture(self, posture_name, percentage_speed):
        """
        Position the virtual robot into a particular posture. The different
        available postures are PepperPosture objects.

        Parameters:
            posture_name - String containing the name of the posture. The
            posture name is not case-sensitive (The available postures
            are Stand or StandInit, StandZero and Crouch)
            percentage_speed - Percentage of the max speed to be used for the
            movement

        Returns:
            Boolean - True if the posture can be applied, False otherwise
        """
        posture_list = [
            PepperVirtual.P_STAND,
            PepperVirtual.P_STAND_INIT,
            PepperVirtual.P_STAND_ZERO,
            PepperVirtual.P_CROUCH]

        for posture in posture_list:
            if posture.isPostureName(posture_name):
                self.setAngles(
                    posture.getPostureJointNames(),
                    posture.getPostureJointValues(),
                    percentage_speed)

                return True

        return False

    def subscribeLaser(self):
        """
        Subscribe to the robot's lasers. Calling this method will launch the
        laser scan process: note that you need the laser scan to be enabled to
        successfully retrieve laser data
        """
        self.laser_manager.subscribe()

    def unsubscribeLaser(self):
        """
        Unsubscribe from the robot's lasers. Calling this method will stop the
        laser scan process
        """
        self.laser_manager.unsubscribe()

    def showLaser(self, display):
        """
        Display debug lines that simulate the laser
        """
        self.laser_manager.showLaser(display)

    def getFrontLaserValue(self):
        """
        Return a list of the front laser value (clockwise)
        """
        return self.laser_manager.getFrontLaserValue()

    def getRightLaserValue(self):
        """
        Return a list of the right laser value (clockwise)
        """
        return self.laser_manager.getRightLaserValue()

    def getLeftLaserValue(self):
        """
        Return a list of the left laser value (clockwise)
        """
        return self.laser_manager.getLeftLaserValue()

    def _mimicHand(self, hand, value, multiplier=0.872665, offset=0):
        """
        Used to propagate a joint value on the fingers attached to the hand.
        The formula used to mimic a joint is the following:

        finger_value = (hand_value * multiplier) + offset

        Parameters:
            hand - String, RHand or LHand
            value - The joint value to be propagated
            multiplier - The multiplier coefficient (0.872665 by default)
            offset - The offset coefficient (0.0 by default)

        Returns:
            finger_names - Names of the finger to be controlled
            finger_values - Values of the fingers to be controlled
        """
        finger_names = list()
        finger_values = list()

        for joint_name in self.joint_dict.keys():
            if (hand[0] + "Finger") in joint_name or\
               (hand[0] + "Thumb") in joint_name:
                finger_names.append(joint_name)
                finger_values.append((value * multiplier) + offset)

        return finger_names, finger_values



#!/usr/bin/env python
# coding: utf-8

import math
import time
import atexit
import weakref
import pybullet
import threading
import numpy as np
from qibullet.link import Link
from qibullet.sensor import Sensor


class CameraResolution:
    """
    Structure for the camera resolutions
    """

    def __init__(self, width, height):
        """
        Constructor

        Parameters:
            width - Width resolution in pixels
            height - Height resolution in pixels
        """
        self.width = width
        self.height = height

    def __eq__(self, resolution):
        """
        Overloading the equal operator

        Parameters:
            resolution - the comparing resolution
        """
        try:
            assert self.width == resolution.width
            assert self.height == resolution.height
            return True

        except AssertionError:
            return False


class Camera(Sensor):
    """
    Class representing a virtual camera
    """
    CAMERA_HANDLES = dict()
    HANDLES_LOCK = threading.Lock()

    K_QQVGA = CameraResolution(160, 120)
    K_QVGA = CameraResolution(320, 240)
    K_VGA = CameraResolution(640, 480)
    K_QQ720p = CameraResolution(320, 180)
    K_Q720p = CameraResolution(640, 360)
    K_720p = CameraResolution(1280, 720)

    def __init__(
            self,
            robot_model,
            camera_id,
            camera_link,
            hfov,
            vfov,
            physicsClientId=0,
            near_plane=0.01,
            far_plane=100):
        """
        Constructor

        Parameters:
            robot_model - The pybullet model of the robot
            camera_id - an int describing the id of the camera
            camera_link - The link (Link type) onto which the camera's attached
            hfov - The value of the horizontal field of view angle (in degrees)
            vfov - The value of the vertical field of view angle (in degrees)
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            near_plane - The near plane distance
            far_plane - The far plane distance
        """
        Sensor.__init__(self, robot_model, physicsClientId)
        self.camera_id = camera_id
        self.camera_link = camera_link
        self.near_plane = near_plane
        self.far_plane = far_plane
        self.frame = None
        self.projection_matrix = None
        self.resolution = None
        self.hfov = None
        self.vfov = None
        self.intrinsic_matrix = None
        self.resolution_lock = threading.Lock()
        self._setFov(hfov, vfov)

    def subscribe(self, resolution):
        """
        Subscribing method for the camera (the FOV has to be specified
        beforehand). This method will launch the frame extraction loop process
        of the camera in a separate thread. This method will return a camera
        handle, needed to retrieve camera frames, get the camera resolution,
        and unsubscribe from it

        Returns:
            handle - The handle of the camera
        """
        # Lets the module process thread die before launching another one, if
        # the same camera calls the subscribing method for a second time
        self._terminateModule()
        self._module_termination = False

        self._setResolution(resolution)

        self.module_process =\
            threading.Thread(target=self._frameExtractionLoop)
        self.module_process.start()
        self._waitForCorrectImageFormat()

        # Add the camera and it's handle to the handles dict
        Camera._addCameraHandle(id(self), self)
        return id(self)

    def unsubscribe(self):
        """
        Method stopping the frame retreival thread for a camera.

        Returns:
            success - Boolean, True if unsubscribed successfully, False
            otherwise
        """
        try:
            self._removeCameraHandle(id(self))
            self._terminateModule()
            self.frame = None
            return True

        except KeyError:
            return False

    def getCameraId(self):
        """
        Returns the id of the camera. WARNING, the id of the camera is not the
        id of the Python object corresponding to the camera, the id value is
        defined by the user and specific to each camera of each robot (
        PepperVirtual.ID_CAMERA_TOP for instance).

        Returns:
            camera_id - The id of the camera
        """
        return self.camera_id

    def getCameraLink(self):
        """
        Returns the link to which the camera is attached, as a qiBullet Link
        object

        Returns:
            camera_link - The link to which the camera is attached
        """
        return self.camera_link

    def getFrame(self):
        """
        Returns the current frame

        Returns:
            frame_copy - a copy of the current frame
        """
        try:
            assert self.frame is not None
            return self.frame.copy()

        except AssertionError:
            return None

    def isActive(self):
        """
        Specifies if the camera is active or not (if a handle exists for the
        current camera)

        Returns:
            is_active - Boolean, True if the camera is subscribed to, False
            otherwise
        """
        if id(self) in Camera._getCameraHandlesDict():
            return True
        else:
            return False

    def getResolution(self):
        """
        Returns the resolution of the camera

        Returns:
            resolution - A CameraResolution object, the resolution of the
            camera
        """
        with self.resolution_lock:
            return self.resolution

    def _setFov(self, hfov, vfov):
        """
        INTERNAL METHOD, sets the field of view of the camera

        Parameters:
            hfov - The value of the horizontal field of view angle (in degrees)
            vfov - The value of the vertical field of view angle (in degrees)
        """
        try:
            assert isinstance(hfov, float) or isinstance(hfov, int)
            assert isinstance(vfov, float) or isinstance(vfov, int)
            self.hfov = hfov
            self.vfov = vfov

        except AssertionError as e:
            print("Cannot set the camera FOV: " + str(e))

    def _setResolution(self, resolution):
        """
        INTERNAL METHOD, sets the resolution for the camera. The fov of the
        camera has to be setted beforehand

        Parameters:
            resolution - The resolution of the camera (CameraResolution type)
        """
        try:
            with self.resolution_lock:
                assert isinstance(resolution, CameraResolution)
                assert self.hfov is not None and self.vfov is not None

                self.resolution = resolution
                self.projection_matrix = pybullet.computeProjectionMatrix(
                    left=-math.tan(math.pi*self.hfov/360.0)*self.near_plane,
                    right=math.tan(math.pi*self.hfov/360.0)*self.near_plane,
                    bottom=-math.tan(math.pi*self.vfov/360.0)*self.near_plane,
                    top=math.tan(math.pi*self.vfov/360.0)*self.near_plane,
                    nearVal=self.near_plane,
                    farVal=self.far_plane,
                    physicsClientId=self.physics_client)

                self.intrinsic_matrix = [
                    self.projection_matrix[0]*self.resolution.width/2.0,
                    0.0,
                    (1-self.projection_matrix[8])*self.resolution.width/2.0,
                    0.0,
                    self.projection_matrix[5]*self.resolution.height/2.0,
                    (1-self.projection_matrix[9])*self.resolution.height/2.0,
                    0.0,
                    0.0,
                    1.0]

        except AssertionError as e:
            raise pybullet.error("Cannot set camera resolution: " + str(e))

    def _getCameraIntrinsics(self):
        """
        INTERNAL METHOD, Returns the intrinsic camera matrix, formatted as:
        K = [fx, 0, cx, 0, fy, cy, 0.0, 0.0, 1.0]. For more details, see
        http://docs.ros.org/kinetic/api/sensor_msgs/html/msg/CameraInfo.html

        Returns:
            intrinsic_matrix - The intrinsic camera matrix formatted as a list,
            or None if the intrinsic camera matrix is None
        """
        with self.resolution_lock:
            try:
                assert self.intrinsic_matrix is not None
                return list(self.intrinsic_matrix)

            except AssertionError:
                print("Cannot get intrinsics, resolution not defined")
                return None

    def _getCameraImage(self):
        """
        INTERNAL METHOD, Computes the OpenGL virtual camera image. The
        resolution and the projection matrix have to be computed before calling
        this method, or it will crash

        Returns:
            camera_image - The camera image of the OpenGL virtual camera
        """
        _, _, _, _, pos_world, q_world = pybullet.getLinkState(
            self.robot_model,
            self.camera_link.getParentIndex(),
            computeForwardKinematics=False,
            physicsClientId=self.physics_client)

        rotation = pybullet.getMatrixFromQuaternion(q_world)
        forward_vector = [rotation[0], rotation[3], rotation[6]]
        up_vector = [rotation[2], rotation[5], rotation[8]]

        camera_target = [
            pos_world[0] + forward_vector[0] * 10,
            pos_world[1] + forward_vector[1] * 10,
            pos_world[2] + forward_vector[2] * 10]

        view_matrix = pybullet.computeViewMatrix(
            pos_world,
            camera_target,
            up_vector,
            physicsClientId=self.physics_client)

        with self.resolution_lock:
            camera_image = pybullet.getCameraImage(
                self.resolution.width,
                self.resolution.height,
                view_matrix,
                self.projection_matrix,
                renderer=pybullet.ER_BULLET_HARDWARE_OPENGL,
                flags=pybullet.ER_NO_SEGMENTATION_MASK,
                physicsClientId=self.physics_client)

        return camera_image

    def _waitForCorrectImageFormat(self):
        """
        INTERNAL METHOD, to be called after the launch of the exctraction
        thread. Blocking method, that will return when the array retrieved from
        the @getCameraImage is not None and corresponds to the current
        resolution
        """
        try:
            assert self.module_process.isAlive()

            while self.getFrame() is None:
                continue

            while True:
                image = self.getFrame()
                if image.shape[1] == self.resolution.width and\
                   image.shape[0] == self.resolution.height:
                    break

        except AssertionError:
            return

    def _frameExtractionLoop(self):
        """
        INTERNAL METHOD, To be specified in each dauhter class, frame
        extraction loop method. This method is threaded. The resolution and the
        FOV have to be specified beforehand
        """
        raise NotImplementedError

    @classmethod
    def _addCameraHandle(cls, handle, camera):
        """
        INTERNAL METHOD, Adds a camera and it's handle to the CAMERA_HANDLES
        dict

        Parameters:
            handle - The camera handle
            camera - The corresponding camera
        """
        with cls.HANDLES_LOCK:
            cls.CAMERA_HANDLES[handle] = camera

    @classmethod
    def _removeCameraHandle(cls, handle):
        """
        INTERNAL METHOD, Remove a camera and it's handle from the
        CAMERA_HANDLES dict. Throws a KeyError if the handle doesn't exist

        Parameters:
            handle - The camera handle
        """
        with cls.HANDLES_LOCK:
            del cls.CAMERA_HANDLES[handle]

    @classmethod
    def _getCameraFromHandle(cls, handle):
        """
        INTERNAL METHOD, Returns the camera associated to the specified handle.
        Throws a KeyError if the handle doesn't exist

        Parameters:
            handle - The camera handle

        Returns:
            camera - The camera object associated to the handle
        """
        with cls.HANDLES_LOCK:
            return cls.CAMERA_HANDLES[handle]

    @classmethod
    def _getCameraHandlesDict(cls):
        """
        INTERNAL METHOD, Returns a COPY of the CAMERA_HANDLES dict

        Returns:
            camera_handles - A copy of the CAMERA_HANDLES dict
        """
        with cls.HANDLES_LOCK:
            return cls.CAMERA_HANDLES.copy()


class CameraRgb(Camera):
    """
    Class representing a virtual rgb camera
    """

    def __init__(
            self,
            robot_model,
            camera_id,
            camera_link,
            hfov,
            vfov,
            physicsClientId=0,
            resolution=Camera.K_QVGA):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            camera_id - an int describing the id of the camera
            camera_link - The link (Link type) onto which the camera's attached
            hfov - The value of the horizontal field of view angle (in degrees)
            vfov - The value of the vertical field of view angle (in degrees)
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            resolution - The resolution of the camera
        """
        Camera.__init__(
            self,
            robot_model,
            camera_id,
            camera_link,
            hfov,
            vfov,
            physicsClientId=physicsClientId)

    def _frameExtractionLoop(self):
        """
        Frame extraction loop, has to be threaded. The resolution and the FOV
        have to be specified beforehand
        """
        rgb_image = np.zeros((
            self.resolution.height,
            self.resolution.width,
            3))

        while not self._module_termination:
            camera_image = self._getCameraImage()

            camera_image = np.reshape(
                camera_image[2],
                (camera_image[1], camera_image[0], 4))

            rgb_image[:, :, 0] =\
                (1 - camera_image[:, :, 3]) * camera_image[:, :, 2] +\
                camera_image[:, :, 3] * camera_image[:, :, 2]

            rgb_image[:, :, 1] =\
                (1 - camera_image[:, :, 3]) * camera_image[:, :, 1] +\
                camera_image[:, :, 3] * camera_image[:, :, 1]

            rgb_image[:, :, 2] =\
                (1 - camera_image[:, :, 3]) * camera_image[:, :, 0] +\
                camera_image[:, :, 3] * camera_image[:, :, 0]

            self.frame = rgb_image.astype(np.uint8)


class CameraDepth(Camera):
    """
    Class representing a virtual depth camera
    """

    def __init__(
            self,
            robot_model,
            camera_id,
            camera_link,
            hfov,
            vfov,
            physicsClientId=0,
            resolution=Camera.K_QVGA):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            camera_id - an int describing the id of the camera
            camera_link - The link (Link type) onto which the camera's attached
            hfov - The value of the horizontal field of view angle (in degrees)
            vfov - The value of the vertical field of view angle (in degrees)
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            resolution - The resolution of the camera
        """
        Camera.__init__(
            self,
            robot_model,
            camera_id,
            camera_link,
            hfov,
            vfov,
            near_plane=0.4,
            far_plane=8,
            physicsClientId=physicsClientId)

    def _frameExtractionLoop(self):
        """
        Frame extraction loop, has to be threaded. The resolution and the FOV
        have to be specified beforehand
        """
        while not self._module_termination:
            camera_image = self._getCameraImage()
            depth_image = np.reshape(
                camera_image[3],
                (camera_image[1], camera_image[0]))

            depth_image = (self.far_plane * self.near_plane) /\
                (self.far_plane - (self.far_plane - self.near_plane) *
                    depth_image)

            depth_image *= 1000
            self.frame = depth_image.astype(np.uint16)


import os
import sys
import site
import platform
import setuptools
from setuptools.command.install import install
from setuptools.command.develop import develop

MESH_LICENSE_AGREEMENT = False


with open("README.md", "r") as fh:
    readme = fh.read()
    readme = readme[:readme.index("<!-- start -->")] +\
        readme[(readme.index("<!-- end -->") + len("<!-- end -->")):]


if "--agree-license" in sys.argv:
    MESH_LICENSE_AGREEMENT = True
    sys.argv.remove("--agree-license")


def get_develop_directory():
    """
    Return the develop directory
    """
    if platform.system() == "Windows":
        return os.path.dirname(os.path.realpath(__file__)) + "\\qibullet"
    else:
        return os.path.dirname(os.path.realpath(__file__)) + "/qibullet"


def get_install_directory():
    """
    Return the installation directory, or None
    """
    if '--user' in sys.argv:
        paths = site.getusersitepackages()
    else:
        paths = site.getsitepackages()

    if isinstance(paths, str):
        paths = [paths]

    for path in paths:
        if platform.system() == "Windows":
            path += "\\qibullet"
        else:
            path += "/qibullet"

        if os.path.exists(path):
            return path

    return None


class RessourceInstallCommand(install):
    def run(self):
        install.run(self)
        install_directory = get_install_directory()

        try:
            assert install_directory is not None

            sys.path.insert(0, install_directory)
            import tools

            if not tools._check_resources_installed():
                tools._install_resources(agreement=MESH_LICENSE_AGREEMENT)

        except AssertionError:
            pass


class RessourceDevelopCommand(develop):
    def run(self):
        develop.run(self)
        develop_directory = get_develop_directory()

        sys.path.insert(0, develop_directory)
        import tools

        if not tools._check_resources_installed():
            tools._install_resources(agreement=MESH_LICENSE_AGREEMENT)


setuptools.setup(
    name="qibullet",
    version="1.4.1",
    author="Maxime Busy, Maxime Caniot",
    author_email="",
    description="Bullet-based simulation for SoftBank Robotics' robots",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/softbankrobotics-research/qibullet",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'pybullet'],
    cmdclass={
        'install': RessourceInstallCommand,
        'develop': RessourceDevelopCommand},
    package_data={"qibullet": [
        "robot_data/*.urdf",
        "robot_data/LICENSE",
        "robot_data/installers/meshes_installer*.pyc",
        "robot_data/meshes.zip"]},
    keywords=[
        'physics simulation',
        'robotics',
        'naoqi',
        'softbank',
        'pepper',
        'nao',
        'romeo',
        'robot'],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        'Topic :: Games/Entertainment :: Simulation',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Framework :: Robot Framework :: Tool'
    ]
)



from qibullet.link import Link
from qibullet.joint import Joint
from qibullet.nao_virtual import NaoVirtual
from qibullet.romeo_virtual import RomeoVirtual
from qibullet.pepper_virtual import PepperVirtual
from qibullet.simulation_manager import SimulationManager
from qibullet.camera import Camera, CameraRgb, CameraDepth, CameraResolution
from qibullet.ros_wrapper import NaoRosWrapper
from qibullet.ros_wrapper import RomeoRosWrapper
from qibullet.ros_wrapper import PepperRosWrapper

name = 'qibullet'
__version__ = "1.4.1"



#!/usr/bin/env python
# coding: utf-8
import os
import sys
import stat
import math
import glob
import shutil
import platform
import pybullet


# Version tag corresponding to the latest version of the additional qibullet
# ressources
RESOURCES_VERSION = "1.4.0"


def getDistance(point_a, point_b):
    [x1, y1, z1] = point_a
    [x2, y2, z2] = point_b
    return int(math.sqrt((x2 - x1)**2 + (y2 - y1)**2 + (z2 - z1)**2) * 100)\
        / 100.0


def getOrientation(theta_a, theta_b):
    return theta_b[-1] - theta_a[-1]


def computeVelocity(acc, vel_min, vel_max, dist_traveled, dist_remained):
    distance_acc = (vel_max * vel_max) / (2 * acc)
    if dist_traveled < distance_acc:
        vel_computed = (vel_max - vel_min) *\
            dist_traveled / distance_acc + vel_min
    if dist_traveled >= distance_acc:
        vel_computed = vel_max
    if dist_remained < distance_acc:
        vel_computed = (vel_max - vel_min) *\
            dist_remained / distance_acc + vel_min
    return vel_computed


def _get_resources_root_folder():  # pragma: no cover
    """
    Returns the path to the resources' root folder (.qibullet folder in the
    user's home). The path will be returned even if the installation folder
    does not yet exist
    """
    if platform.system() == "Windows":
        return os.path.expanduser("~") + "\\.qibullet"
    else:
        return os.path.expanduser("~") + "/.qibullet"


def _get_resources_folder():  # pragma: no cover
    """
    Returns the path to the resources folder specific to the current qibullet
    version (.qibullet/resources_version folder in the user's home). The path
    will be returned even if the installation folder does not yet exists
    """
    if platform.system() == "Windows":
        return _get_resources_root_folder() + "\\" + RESOURCES_VERSION
    else:
        return _get_resources_root_folder() + "/" + RESOURCES_VERSION


def _install_resources(agreement=False):  # pragma: no cover
    """
    Extracts the robot meshes and install the urdfs and the meshes, using the
    provided installers. The resources will be installed in the user's home
    folder (under the .qibullet folder)
    """
    # If the install folder already exists, remove it
    if not _uninstall_resources():
        print(
            "Cannot install the ressources, try to manually remove the " +
            _get_resources_root_folder() + " folder first.")
        return

    resources_folder = _get_resources_folder()

    # Displaying the qiBullet version corresponding to the extra resources
    print("\nInstalling resources for qiBullet")

    # Ask for user feedback before installing everything.
    try:
        assert not agreement

        if sys.version_info > (3, 0):
            answer = input(
                "\nThe robot meshes and URDFs will be installed in the " +
                resources_folder + " folder. You will need to agree to"
                " the meshes license in order to be able to install them."
                " Continue the installation (y/n)? ")
        else:
            answer = raw_input(
                "\nThe robot meshes and URDFs will be installed in the " +
                resources_folder + " folder. You will need to agree to"
                " the meshes license in order to be able to install them."
                " Continue the installation (y/n)? ")

    except AssertionError:
        answer = "y"

    if answer.lower() == "y":
        print(
            "Installing the meshes and URDFs in the " + resources_folder +
            " folder...")
    else:
        print("The meshes and URDFs won't be installed.")
        return

    # Create the resources root folder
    os.mkdir(_get_resources_root_folder())
    # Create the version specific resources folder
    os.mkdir(resources_folder)

    # Fetch the correct installer and extract the robot meshes in the install
    # folder
    if platform.system() == "Windows":
        data_folder = os.path.dirname(os.path.realpath(__file__)) +\
            "\\robot_data\\"
    else:
        data_folder = os.path.dirname(os.path.realpath(__file__)) +\
            "/robot_data/"

    sys.path.insert(0, data_folder + "installers")
    major = sys.version_info[0]
    minor = sys.version_info[1]
    print("Python " + str(major) + "." + str(minor) + " detected")

    if major == 3:
        if minor == 5:
            import meshes_installer_35 as meshes_installer
        elif minor == 6:
            import meshes_installer_36 as meshes_installer
        elif minor == 7:
            import meshes_installer_37 as meshes_installer
        elif minor == 8:
            import meshes_installer_38 as meshes_installer
        else:
            print("Uncompatible version of Python 3")
            return
    elif major == 2:
        if minor == 7:
            import meshes_installer_27 as meshes_installer
        else:
            print("Uncompatible version of Python 2")
            return
    else:
        print("Uncompatible Python version")
        return

    if meshes_installer._install_meshes(resources_folder, agreement=agreement):
        print("Resources correctly extracted")
    else:
        print("Could not extract the resources")
        return

    # Install the robot URDFs in the install folder
    print("Installing the robot URDFs...")

    for urdf in glob.glob(data_folder + "*.urdf"):
        shutil.copy2(urdf, resources_folder)

    # Grant writing permissions on the ressources
    if platform.system() != "Windows":
        permissions = stat.S_IWOTH | stat.S_IWGRP
        os.chmod(
            _get_resources_root_folder(),
            permissions | os.stat(_get_resources_root_folder()).st_mode)

        for root, folders, files in os.walk(_get_resources_root_folder()):
            for folder in folders:
                folder_path = os.path.join(root, folder)
                os.chmod(
                    folder_path,
                    permissions | os.stat(folder_path).st_mode)

            for ressource_file in files:
                file_path = os.path.join(root, ressource_file)
                os.chmod(file_path, permissions | os.stat(file_path).st_mode)

    print(
        "(To remove the installed resources, use the _uninstall_resources "
        "method of qibullet.tools, or remove the folder manually)")
    print("Installation done, resources in " + resources_folder)


def _uninstall_resources():  # pragma: no cover
    """
    Uninstall the robot meshes and the urdfs from the user's home folder
    (removing the .qibullet folder in the user's home). Will return True if the
    .qibullet folder doesn't exit in the user's home anymore
    """
    if os.path.exists(_get_resources_root_folder()):
        try:
            shutil.rmtree(_get_resources_root_folder())

        except OSError:
            return False

    return True


def _check_resources_installed():  # pragma: no cover
    """
    Checks if the resources (URDFs and robot meshes) are install in the user's
    home folder (in the .qibullet folder)

    Returns:
        installed - boolean, True if the meshes are installed, False otherwise
    """
    install_folder = os.path.dirname(os.path.realpath(__file__))

    try:
        assert os.path.exists(_get_resources_root_folder())

    except AssertionError:
        print("\nThe qibullet ressources are not yet installed.")
        return False

    try:
        assert os.path.exists(_get_resources_folder())

    except AssertionError:
        print("\nThe qibullet ressources are not up to date.")
        return False

    try:
        if platform.system() == "Windows":
            assert os.path.exists(_get_resources_folder() + "\\nao.urdf")
            assert os.path.exists(_get_resources_folder() + "\\romeo.urdf")
            assert os.path.exists(_get_resources_folder() + "\\pepper.urdf")
            assert os.path.exists(_get_resources_folder() + "\\meshes")
        else:
            assert os.path.exists(_get_resources_folder() + "/nao.urdf")
            assert os.path.exists(_get_resources_folder() + "/romeo.urdf")
            assert os.path.exists(_get_resources_folder() + "/pepper.urdf")
            assert os.path.exists(_get_resources_folder() + "/meshes")

        return True

    except AssertionError:
        print("\nThe qibullet ressources are up to date but seem incomplete.")
        return False
